""" Helper functions for cci_plot.py.
"""

import sys
import math
import numpy as np
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.patches import Arc, Wedge
from mpl_toolkits.axes_grid1 import make_axes_locatable

from matplotlib.path import Path
import matplotlib.patches as patches
import matplotlib.colors as plt_colors
import matplotlib.cm as cm

from ..tools.microenv.cci.het import get_edges

from anndata import AnnData

# Helper functions for overview plots of the LRs.


def lr_scatter(
    data,
    feature,
    highlight_lrs=None,
    show_text=True,
    n_top=50,
    color="gold",
    alpha=0.5,
    lr_text_fp=None,
    axis_text_fp=None,
    ax=None,
    show=True,
    max_text=100,
    highlight_color="red",
    figsize: tuple = None,
    show_all: bool = False,
):
    """General plotting of the LR features."""
    highlight = type(highlight_lrs) != type(None)
    if not highlight:
        show_text = show_text if n_top <= max_text else False
    else:
        highlight_lrs = highlight_lrs[0:max_text]

    lr_df = data.uns["lr_summary"]
    lrs = lr_df.index.values.astype(str)[0:n_top]
    lr_features = data.uns["lrfeatures"]
    lr_df = pd.concat([lr_df, lr_features], axis=1).loc[lrs, :]
    if feature not in lr_df.columns:
        raise Exception(f"Inputted {feature}; must be one of " f"{list(lr_df.columns)}")

    rot = 90 if feature != "n_spots_sig" else 70

    n_spots = lr_df.loc[:, feature].values[0:n_top]

    return rank_scatter(
        lrs,
        n_spots,
        y_label=feature,
        x_label="LR Rank (n_sig_spots)",
        figsize=figsize,
        highlight_items=highlight_lrs,
        show_text=show_text,
        color=color,
        alpha=alpha,
        lr_text_fp=lr_text_fp,
        axis_text_fp=axis_text_fp,
        ax=ax,
        show=show,
        highlight_color=highlight_color,
        rot=rot,
        pad=0,
        show_all=show_all,
    )
    # ranks = np.array(list(range(len(n_spots))))
    #
    # if type(lr_text_fp)==type(None):
    #     lr_text_fp = {'weight': 'bold', 'size': 8}
    # if type(axis_text_fp)==type(None):
    #     axis_text_fp = {'weight': 'bold', 'size': 12}
    #
    # if type(ax)==type(None):
    #     width = (7.5 / 50) * n_top if show_text and not highlight else 7.5
    #     if width > 20:
    #         width = 20
    #     fig, ax = plt.subplots(figsize=(width, 4))
    #
    # # Plotting the points #
    # ax.scatter(ranks, n_spots, alpha=alpha, c=color)
    #
    # if show_text:
    #     if highlight:
    #         ranks = ranks[[np.where(lrs==lr)[0][0] for lr in highlight_lrs]]
    #         ax.scatter(ranks, n_spots[ranks], alpha=alpha, c=highlight_color)
    #
    #     for i in ranks:
    #         ax.text(i-.2, n_spots[i], lrs[i], rotation=rot, fontdict=lr_text_fp)
    #
    # ax.spines['top'].set_visible(False)
    # ax.spines['right'].set_visible(False)
    # ax.set_xlabel('LR Rank', axis_text_fp)
    # ax.set_ylabel(feature, axis_text_fp)
    #
    # if show:
    #     plt.show()
    # else:
    #     return ax


def rank_scatter(
    items,
    y,
    y_label: str = "",
    x_label: str = "",
    highlight_items=None,
    show_text=True,
    color="gold",
    alpha=0.5,
    lr_text_fp=None,
    axis_text_fp=None,
    ax=None,
    show=True,
    highlight_color="red",
    rot: float = 90,
    point_sizes: np.array = None,
    pad=0.2,
    figsize=None,
    width_ratio=7.5 / 50,
    height=4,
    point_size_name="Sizes",
    point_size_exp=2,
    show_all: bool = False,
):
    """General plotting function for showing ranked list of items."""
    ranks = np.array(list(range(len(items))))

    highlight = type(highlight_items) != type(None)
    if type(lr_text_fp) == type(None):
        lr_text_fp = {"weight": "bold", "size": 8}
    if type(axis_text_fp) == type(None):
        axis_text_fp = {"weight": "bold", "size": 12}

    if type(ax) == type(None):
        if type(figsize) == type(None):
            width = width_ratio * len(ranks) if show_text and not highlight else 7.5
            if width > 20:
                width = 20
            figsize = (width, height)
        fig, ax = plt.subplots(figsize=figsize)

    # Plotting the points #
    scatter = ax.scatter(
        ranks,
        y,
        alpha=alpha,
        c=color,
        s=None if type(point_sizes) == type(None) else point_sizes ** point_size_exp,
        edgecolors="none",
    )
    y_min, y_max = ax.get_ylim()
    y_max = y_max + y_max * pad
    ax.set_ylim(y_min, y_max)
    if type(point_sizes) != type(None):
        # produce a legend with a cross section of sizes from the scatter
        handles, labels = scatter.legend_elements(prop="sizes", alpha=0.6, num=4)
        [handle.set_markeredgecolor("none") for handle in handles]
        starts = [label.find("{") for label in labels]
        ends = [label.find("}") + 1 for label in labels]
        sizes = [
            float(label[(starts[i] + 1) : (ends[i] - 1)])
            for i, label in enumerate(labels)
        ]
        counts = [int(size ** (1 / point_size_exp)) for size in sizes]
        labels2 = [
            label.replace(label[(starts[i]) : (ends[i])], "{" + str(counts[i]) + "}")
            for i, label in enumerate(labels)
        ]
        legend2 = ax.legend(
            handles,
            labels2,
            frameon=False,
            # bbox_to_anchor=(0.1, 0.05, 1., 1.),
            handletextpad=1.6,
            loc="upper right",
            title=point_size_name,
        )

    if show_text:
        if highlight:
            ranks_ = ranks[[np.where(items == item)[0][0] for item in highlight_items]]
            ax.scatter(
                ranks_,
                y[ranks_],
                alpha=alpha,
                c=highlight_color,
                s=None
                if type(point_sizes) == type(None)
                else (point_sizes[ranks_] ** point_size_exp),
                edgecolors=color,
            )
            ranks = ranks_ if not show_all else ranks

        for i in ranks:
            ax.text(i - 0.2, y[i], items[i], rotation=rot, fontdict=lr_text_fp)

    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.set_xlabel(x_label, axis_text_fp)
    ax.set_ylabel(y_label, axis_text_fp)

    if show:
        plt.show()
    else:
        return ax


# Helper functions for lr_plot


def add_arrows(
    adata: AnnData,
    l_expr: np.array,
    r_expr: np.array,
    min_expr: float,
    sig_bool: np.array,
    fig,
    ax: Axes,
    use_label: str,
    int_df: pd.DataFrame,
    head_width=4,
    width=0.001,
    arrow_cmap=None,
    arrow_vmax=None,
):
    """ Adds arrows to the current plot for significant spots to neighbours \
        which is interacting with.
        Parameters
        ----------
        adata: AnnData          The anndata object.
        L_bool: np.array
        Returns
        -------
        counts: int   Total number of interactions satisfying the conditions, \
                      or np.array<set> if return_edges=True, where each set is \
                      an edge, only returns unique edges.
    """

    library_id = list(adata.uns["spatial"].keys())[0]
    res = adata.uns["spatial"][library_id]["use_quality"]
    scale_factor = adata.uns["spatial"][library_id]["scalefactors"][
        f"tissue_{res}_scalef"
    ]
    # scale_factor = 1
    # scale_factor = 1 # scale factor pretty much isn't used since always parsed as 1
    # in the base plotting function class.
    # Reason why is because scale_factor refers to scaling the
    # image to match the spot spatial coordinates, not the
    # the spots to match the image coordinates!!!

    L_bool = l_expr > min_expr
    R_bool = r_expr > min_expr

    # Getting the edges, from sig-L->R and sig-R<-L #
    forward_edges, reverse_edges = get_edges(adata, L_bool, R_bool, sig_bool)

    # If int_df specified, means need to subset to CCIs which are significant #
    if type(int_df) != type(None):
        spot_bcs = adata.obs_names.values.astype(str)
        spot_labels = adata.obs[use_label].values.astype(str)
        label_set = int_df.index.values.astype(str)
        interact_bool = int_df.values > 0

        # Subsetting to only significant CCI #
        edges_sub = [[], []]  # forward, reverse
        # ints_2 = np.zeros(int_df.shape) # Just for debugging make sure edge
        # list re-capitulates edge-counts.
        for i, edges in enumerate([forward_edges, reverse_edges]):
            for j, edge in enumerate(edges):
                k_ = [0, 1] if i == 0 else [1, 0]
                celltype0 = np.where(label_set == spot_labels[spot_bcs == edge[k_[0]]])[
                    0
                ][0]
                celltype1 = np.where(label_set == spot_labels[spot_bcs == edge[k_[1]]])[
                    0
                ][0]
                celltypes = np.array([celltype0, celltype1])
                # For debugging purposes, used to find indexing bug #
                # print(label_set[celltypes[k_[0]]],
                #       label_set[celltypes[k_[1]]])
                # print( celltypes[k_[0]], celltypes[k_[1]] )
                # print( interact_bool[celltypes[k_[0]], celltypes[k_[1]]] )
                if interact_bool[celltypes[k_[0]], celltypes[k_[1]]]:
                    edges_sub[i].append(edge)
                    # ints_2[celltypes[k_[0]], celltypes[k_[1]]] += 1

        forward_edges, reverse_edges = edges_sub

    # If cmap specified, colour arrows by average LR expression on edge #
    if type(arrow_cmap) != type(None):
        edges_means = [[], []]
        all_means = []
        for i, edges in enumerate([forward_edges, reverse_edges]):
            for j, edge in enumerate(edges):
                edge0_bool = spot_bcs == edge[0]
                edge1_bool = spot_bcs == edge[1]
                l_expr0 = l_expr[edge0_bool]
                l_expr1 = l_expr[edge1_bool]
                r_expr0 = r_expr[edge0_bool]
                r_expr1 = r_expr[edge1_bool]
                mean_expr = np.mean([l_expr0, l_expr1, r_expr0, r_expr1])
                edges_means[i].append(mean_expr)
                all_means.append(mean_expr)

        # Determining the color maps #
        arrow_vmax = np.max(all_means) if type(arrow_vmax) == type(None) else arrow_vmax
        cmap = plt.get_cmap(arrow_cmap)
        c_norm = plt_colors.Normalize(vmin=0, vmax=arrow_vmax)
        scalar_map = cm.ScalarMappable(norm=c_norm, cmap=cmap)

        # Determining the edge colors #
        edges_colors = [[], []]
        for i, edges in enumerate([forward_edges, reverse_edges]):
            for j, edge in enumerate(edges):
                color_val = scalar_map.to_rgba(edges_means[i][j])
                edges_colors[i].append(color_val)

        # Need to add new axes #
        # xlims = ax.get_xlim()
        # ylims = ax.get_ylim()
        # left, bottom = xlims[0]*0.025, ylims[0]*0.17
        # axc = fig.add_axes([0, 0, 0.28, 0.015])
        divider = make_axes_locatable(ax)
        cax = divider.new_vertical(size="2%", pad=0.1, pack_start=True)
        axc = fig.add_axes(cax)

    else:
        edges_colors = [None, None]

    # Now performing the plotting #
    # The arrows #
    # Now converting the edges to coordinates #
    add_arrows_by_edges(
        ax,
        adata,
        forward_edges,
        scale_factor,
        head_width,
        width,
        edge_colors=edges_colors[0],
    )
    add_arrows_by_edges(
        ax,
        adata,
        reverse_edges,
        scale_factor,
        head_width,
        width,
        forward=False,
        edge_colors=edges_colors[1],
    )
    # Adding the color map #
    if type(arrow_cmap) != type(None):
        cb1 = matplotlib.colorbar.ColorbarBase(
            axc, cmap=cmap, norm=c_norm, orientation="horizontal"
        )


def add_arrows_by_edges(
    ax,
    adata,
    edges,
    scale_factor,
    head_width,
    width,
    forward=True,
    edge_colors=None,
    axc=None,
):
    """Adds the arrows using an edge list."""
    for i, edge in enumerate(edges):
        # cols = ["imagecol", "imagerow"]
        if forward:
            edge0, edge1 = edge
        else:
            edge0, edge1 = edge[::-1]

        # Arrow details #
        # x1, y1 = adata.obs.loc[edge0, cols].values.astype(float) * scale_factor
        # x2, y2 = adata.obs.loc[edge1, cols].values.astype(float) * scale_factor
        edge0_index = np.where(adata.obs_names.values == edge0)[0][0]
        edge1_index = np.where(adata.obs_names.values == edge1)[0][0]
        x1, y1 = adata.obsm["spatial"][edge0_index, :] * scale_factor
        x2, y2 = adata.obsm["spatial"][edge1_index, :] * scale_factor
        dx, dy = (x2 - x1) * 0.75, (y2 - y1) * 0.75
        arrow_color = "k" if type(edge_colors) == type(None) else edge_colors[i]

        ax.arrow(
            x1,
            y1,
            dx,
            dy,
            head_width=head_width,
            width=width,
            linewidth=0.01,
            facecolor=arrow_color,
        )


# Helper functions for cci_map


def get_int_df(adata, lr, use_label, sig_interactions, title):
    """Retrieves the relevant interaction count matrix."""
    no_title = type(title) == type(None)
    if type(lr) == type(None):  # No LR inputted, so just use all
        int_df = (
            adata.uns[f"lr_cci_{use_label}"]
            if sig_interactions
            else adata.uns[f"lr_cci_raw_{use_label}"]
        )
        title = "Cell-Cell LR Interactions" if no_title else title
    else:
        int_df = (
            adata.uns[f"per_lr_cci_{use_label}"][lr]
            if sig_interactions
            else adata.uns[f"per_lr_cci_raw_{use_label}"][lr]
        )
        title = f"Cell-Cell {lr} interactions" if no_title else title

    return int_df, title


def create_flat_df(int_df):
    """Reformats a dataframe representing interactions to a flat format."""
    n_rows = int_df.shape[0] * int_df.shape[1]
    flat_df = pd.DataFrame(index=list(range(n_rows)), columns=["x", "y", "value"])
    row_i = 0
    for i, cell_typei in enumerate(int_df.index.values):
        for j, cell_typej in enumerate(int_df.columns.values):
            flat_df.iloc[row_i, :] = [cell_typei, cell_typej, int_df.values[i, j]]
            row_i += 1

    return flat_df


def _box_map(x, y, size, ax=None, figsize=(6.48, 4.8), cmap=None, square_scaler=700):
    """Main underlying helper function for generating the heatmaps."""
    if type(cmap) == type(None):
        cmap = "Spectral_r"

    if type(ax) == type(None):
        fig, ax = plt.subplots(figsize=figsize)

    # Mapping from column names to integer coordinates
    x_labels = list(x.values)  # [v for v in sorted(x.unique())]
    y_labels = list(y.values)  # [v for v in sorted(y.unique())]
    x_to_num = {p[1]: p[0] for p in enumerate(x_labels)}
    y_to_num = {p[1]: p[0] for p in enumerate(y_labels)}

    out = ax.scatter(
        x=x.map(x_to_num),  # Use mapping for x
        y=y.map(y_to_num),  # Use mapping for y
        s=size / sum(size) * square_scaler,
        c=size,
        cmap=cmap,
        # Vector of square sizes, proportional to size parameter
        marker="s",  # Use square as scatterplot marker
    )
    out.set_array(size.values.astype(int))
    out.set_clim(min(size), max(size))
    cbar = plt.colorbar(out)
    cbar.ax.get_yaxis().labelpad = 15
    cbar.ax.set_ylabel("# of interactions", rotation=270)

    # Show column labels on the axes
    ax.set_xticks([x_to_num[v] for v in x_labels])
    ax.set_xticklabels(x_labels, rotation=45, horizontalalignment="right")
    ax.set_yticks([y_to_num[v] for v in y_labels])
    ax.set_yticklabels(y_labels)

    return ax


# Matplotlib chord diagram, adapted & refactored from:
# https://github.com/fengwangPhysics/matplotlib-chord-diagram/blob/master/matplotlib-chord.py

###################
# chord diagram
LW = 0.3


def polar2xy(r, theta):
    return np.array([r * np.cos(theta), r * np.sin(theta)])


def hex2rgb(c):
    return tuple(int(c[i : i + 2], 16) / 256.0 for i in (1, 3, 5))


def IdeogramArc(
    start=0, end=60, radius=1.0, width=0.2, ax=None, color=(1, 0, 0), curve_steps=1
):
    # start, end should be in [0, 360)
    if start > end:
        start, end = end, start
    start *= np.pi / 180.0
    end *= np.pi / 180.0
    # optimal distance to the control points
    # https://stackoverflow.com/questions/1734745/how-to-create-circle-with-b%C3%A9zier-curves
    opt = 4.0 / 3.0 * np.tan((end - start) / 4.0) * radius
    inner = radius * (1 - width)

    # Defines descrete points along the arc #
    # The outter part of the arc #

    val = 0.5
    verts_upper_start = [polar2xy(radius, start)]
    verts_upper_curve = [
        polar2xy(radius, start)
        + polar2xy(opt, start + (np.pi * ((val * i) / curve_steps)))
        for i in range(1, curve_steps + 1)
    ]
    verts_upper_curve += [
        polar2xy(radius, end) + polar2xy(opt, end - (np.pi * ((val * i) / curve_steps)))
        for i in range(1, curve_steps + 1)
    ]
    verts_upper = verts_upper_start + verts_upper_curve + [polar2xy(radius, end)]

    verts_inner_start = [polar2xy(inner, end)]
    verts_inner_curve = [
        polar2xy(inner, end)
        + polar2xy(opt * (1 - width), end - (np.pi * ((val * i) / curve_steps)))
        for i in range(1, curve_steps + 1)
    ]
    verts_inner_curve += [
        polar2xy(inner, start)
        + polar2xy(opt * (1 - width), start + (np.pi * ((val * i) / curve_steps)))
        for i in range(1, curve_steps + 1)
    ]
    verts_inner = (
        verts_inner_start
        + verts_inner_curve
        + [polar2xy(inner, start), polar2xy(radius, start)]
    )

    verts = verts_upper + verts_inner

    codes = (
        [Path.MOVETO]
        + [Path.CURVE4] * curve_steps * 2
        + [Path.CURVE4, Path.LINETO]
        + [Path.CURVE4] * curve_steps * 2
        + [
            Path.CURVE4,
            Path.CLOSEPOLY,
        ]
    )

    if ax == None:
        return verts, codes
    else:
        path = Path(verts, codes)
        patch = patches.PathPatch(
            path, facecolor=color + (0.5,), edgecolor=color + (0.4,), lw=LW
        )
        ax.add_patch(patch)


def ChordArc(
    start1=0,
    end1=60,
    start2=180,
    end2=240,
    radius=1.0,
    chordwidth=0.7,
    ax=None,
    color=(1, 0, 0),
):
    # start, end should be in [0, 360)
    if start1 > end1:
        start1, end1 = end1, start1
    if start2 > end2:
        start2, end2 = end2, start2
    start1 *= np.pi / 180.0
    end1 *= np.pi / 180.0
    start2 *= np.pi / 180.0
    end2 *= np.pi / 180.0
    opt1 = 4.0 / 3.0 * np.tan((end1 - start1) / 4.0) * radius
    opt2 = 4.0 / 3.0 * np.tan((end2 - start2) / 4.0) * radius
    rchord = radius * (1 - chordwidth)
    verts = [
        polar2xy(radius, start1),
        polar2xy(radius, start1) + polar2xy(opt1, start1 + 0.5 * np.pi),
        polar2xy(radius, end1) + polar2xy(opt1, end1 - 0.5 * np.pi),
        polar2xy(radius, end1),
        polar2xy(rchord, end1),
        polar2xy(rchord, start2),
        polar2xy(radius, start2),
        polar2xy(radius, start2) + polar2xy(opt2, start2 + 0.5 * np.pi),
        polar2xy(radius, end2) + polar2xy(opt2, end2 - 0.5 * np.pi),
        polar2xy(radius, end2),
        polar2xy(rchord, end2),
        polar2xy(rchord, start1),
        polar2xy(radius, start1),
    ]

    codes = [
        Path.MOVETO,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
    ]

    if ax == None:
        return verts, codes
    else:
        path = Path(verts, codes)
        patch = patches.PathPatch(
            path, facecolor=color + (0.5,), edgecolor=color + (0.4,), lw=LW
        )
        ax.add_patch(patch)


def selfChordArc(start=0, end=60, radius=1.0, chordwidth=0.7, ax=None, color=(1, 0, 0)):
    # start, end should be in [0, 360)
    if start > end:
        start, end = end, start
    start *= np.pi / 180.0
    end *= np.pi / 180.0
    opt = 4.0 / 3.0 * np.tan((end - start) / 4.0) * radius
    rchord = radius * (1 - chordwidth)
    verts = [
        polar2xy(radius, start),
        polar2xy(radius, start) + polar2xy(opt, start + 0.5 * np.pi),
        polar2xy(radius, end) + polar2xy(opt, end - 0.5 * np.pi),
        polar2xy(radius, end),
        polar2xy(rchord, end),
        polar2xy(rchord, start),
        polar2xy(radius, start),
    ]

    codes = [
        Path.MOVETO,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
        Path.CURVE4,
    ]

    if ax == None:
        return verts, codes
    else:
        path = Path(verts, codes)
        patch = patches.PathPatch(
            path, facecolor=color + (0.5,), edgecolor=color + (0.4,), lw=LW
        )
        ax.add_patch(patch)


def chordDiagram(X, ax, colors=None, width=0.1, pad=2, chordwidth=0.7, lim=1.1):
    """Plot a chord diagram
    Parameters
    ----------
    X :
        flux data, X[i, j] is the flux from i to j
    ax :
        matplotlib `axes` to show the plot
    colors : optional
        user defined colors in rgb format. Use function hex2rgb() to convert hex color to rgb color. Default: d3.js category10
    width : optional
        width/thickness of the ideogram arc
    pad : optional
        gap pad between two neighboring ideogram arcs, unit: degree, default: 2 degree
    chordwidth : optional
        position of the control points for the chords, controlling the shape of the chords
    """
    # X[i, j]:  i -> j
    x = X.sum(axis=1)  # sum over rows
    ax.set_xlim(-lim, lim)
    ax.set_ylim(-lim, lim)
    diam = 1.8

    if colors is None:
        # use d3.js category10 https://github.com/d3/d3-3.x-api-reference/blob/master/Ordinal-Scales.md#category10
        colors = [
            "#1f77b4",
            "#ff7f0e",
            "#2ca02c",
            "#d62728",
            "#9467bd",
            "#8c564b",
            "#e377c2",
            "#7f7f7f",
            "#bcbd22",
            "#17becf",
        ]
        if len(x) > 10:
            print("x is too large! Use x smaller than 10")
    if type(colors[0]) == str:
        colors = [hex2rgb(colors[i]) for i in range(len(x))]

    # find position for each start and end
    y = x / np.sum(x).astype(float) * (360 - pad * len(x))

    pos = {}
    arc = []
    nodePos = []
    start = 0
    for i in range(len(x)):
        end = start + y[i]
        arc.append((start, end))
        angle = 0.5 * (start + end)
        # print(start, end, angle)
        if -30 <= angle <= 210:
            angle -= 90
        else:
            angle -= 270
        nodePos.append(
            tuple(
                polar2xy((diam / 2) + diam * 0.05, 0.5 * (start + end) * np.pi / 180.0)
            )
            + (angle,)
        )
        z = (X[i, :] / x[i].astype(float)) * (end - start)
        ids = np.argsort(z)
        z0 = start
        for j in ids:
            pos[(i, j)] = (z0, z0 + z[j])
            z0 += z[j]
        start = end + pad

    for i in range(len(x)):
        start, end = arc[i]
        # This draws the outter ring #
        # IdeogramArc(start=start, end=end, radius=1.0, ax=ax,
        #            color=colors[i], width=width)
        a = Arc((0, 0), diam, diam, 0, start, end, color=colors[i], lw=10)
        ax.add_patch(a)
        start, end = pos[(i, i)]
        # This draws the paths to itself #
        if end - start < 180:  # Indicates this method will work fine !
            selfChordArc(
                start,
                end,
                radius=1.0 - width,
                color=colors[i],
                chordwidth=chordwidth * 0.7,
                ax=ax,
            )
        else:  # Need to use a wedge because the arch distorts past 180-degrees
            path = Wedge(0, diam / 2, start, end, color=colors[i] + (0.5,))
            ax.add_patch(path)
        for j in range(i):
            if X[i, j] == 0 and X[j, i] == 0:  # don't draw anything for no interaction
                continue
            color = colors[i]
            if X[i, j] > X[j, i]:  # Color by the dominant signal #
                color = colors[j]
            start1, end1 = pos[(i, j)]
            start2, end2 = pos[(j, i)]
            ChordArc(
                start1,
                end1,
                start2,
                end2,
                radius=1.0 - width,
                color=color,
                chordwidth=chordwidth,
                ax=ax,
            )

    return nodePos



import numpy as np
import pandas as pd
from anndata import AnnData
import scipy.spatial as spatial
from numba.typed import List
from numba import njit, jit

from stlearn.tools.microenv.cci.het_helpers import (
    edge_core,
    get_between_spot_edge_array,
    get_data_for_counting,
    add_unique_edges,
    get_neighbourhoods,
    init_edge_list,
)


def count(
    adata: AnnData,
    use_label: str = None,
    use_het: str = "cci_het",
    verbose: bool = True,
    distance: float = None,
) -> AnnData:
    """Count the cell type densities
    Parameters
    ----------
    adata: AnnData          The data object including the cell types to count
    use_label:         The cell type results to use in counting
    use_het:                The stoarge place for result
    distance: int           Distance to determine the neighbours (default is the nearest neighbour), distance=0 means within spot

    Returns
    -------
    adata: AnnData          With the counts of specified clusters in nearby spots stored as adata.uns['het']
    """

    library_id = list(adata.uns["spatial"].keys())[0]
    # between spot
    if distance != 0:
        # automatically calculate distance if not given, won't overwrite distance=0 which is within-spot
        if not distance:
            # calculate default neighbour distance
            scalefactors = next(iter(adata.uns["spatial"].values()))["scalefactors"]
            distance = (
                scalefactors["spot_diameter_fullres"]
                * scalefactors[
                    "tissue_"
                    + adata.uns["spatial"][library_id]["use_quality"]
                    + "_scalef"
                ]
                * 2
            )

        counts_ct = pd.DataFrame(0, adata.obs_names, ["CT"])

        # get neighbour spots for each spot
        coor = adata.obs[["imagerow", "imagecol"]]
        point_tree = spatial.cKDTree(coor)
        neighbours = []
        for spot in adata.obs_names:
            n_index = point_tree.query_ball_point(
                np.array(
                    [adata.obs["imagerow"].loc[spot], adata.obs["imagecol"].loc[spot]]
                ),
                distance,
            )
            neighbours = [item for item in adata.obs_names[n_index]]
            counts_ct.loc[spot] = (
                (adata.uns[use_label].loc[neighbours] > 0.2).sum() > 0
            ).sum()
        adata.obsm[use_het] = counts_ct["CT"].values

    # within spot
    else:
        # count the cell types with prob > 0.2 in the result of label transfer
        adata.obsm[use_het] = (adata.uns[use_label] > 0.2).sum(axis=1)

    if verbose:
        print(
            "Counts for cluster (cell type) diversity stored into adata.uns['"
            + use_het
            + "']"
        )

    return adata


def get_edges(adata: AnnData, L_bool: np.array, R_bool: np.array, sig_bool: np.array):
    """Gets a list edges representing significant interactions.

    Parameters
    ----------
    adata: AnnData
    L_bool: np.array<bool>  len(L_bool)==len(adata), True if ligand expressed in that spot.
    R_bool: np.array<bool>  len(R_bool)==len(adata), True if receptor expressed in that spot.
    sig_bool np.array<bool>:   len(sig_bool)==len(adata), True if spot has significant LR interactions.
    Returns
    -------
    edge_list_unique:   list<list<str>> Either a list of tuples (directed), or
                        list of sets (undirected), indicating unique significant
                        interactions between spots.
    """
    # Getting the neighbourhoods #
    neighbours, neighbourhood_bcs, neighbourhood_indices = get_neighbourhoods(adata)

    # Getting the edges to draw in-between #
    L_spot_indices = np.where(np.logical_and(L_bool, sig_bool))[0]
    R_spot_indices = np.where(np.logical_and(R_bool, sig_bool))[0]

    # To keep the get_between_spot_edge_array function happy #
    cell_data = np.ones((1, len(sig_bool)))[0, :].astype(np.float_)

    # Retrieving the edges #
    gene_bools = [R_bool, L_bool]
    edge_list = []
    for i, spot_indices in enumerate([L_spot_indices, R_spot_indices]):
        # Subsetting to the relevant neighbourhoods #
        neigh_bcs = List()
        neigh_indices = List()
        for j in spot_indices:
            neigh_bcs.append(neighbourhood_bcs[j])
            neigh_indices.append(neighbourhood_indices[j])

        # Getting the edges in this neighbourhood #
        edges = init_edge_list(neighbourhood_bcs)  # Note this has 1 pseudo edge
        if len(spot_indices) != 0:
            get_between_spot_edge_array(
                edges, neigh_bcs, neigh_indices, gene_bools[i], cell_data
            )
        edge_list.append(edges[1:])

    return edge_list


def count_interactions(
    adata,
    all_set,
    mix_mode,
    use_label,
    sig_bool,
    gene1_bool,
    gene2_bool,
    tissue_types=None,
    cell_type_props=None,
    cell_prop_cutoff=0.2,
    trans_dir=True,
):
    """Counts the interactions."""
    # Getting minimal information necessary for the counting #
    (
        spot_bcs,
        cell_data,
        neighbourhood_bcs,
        neighbourhood_indices,
    ) = get_data_for_counting(adata, use_label, mix_mode, all_set)

    # if trans_dir, rows are transmitter cell, cols receiver, otherwise reverse.
    int_matrix = np.zeros((len(all_set), len(all_set)), dtype=int)
    for i, cell_A in enumerate(all_set):  # transmitter if trans_dir else reciever
        # Determining which spots have cell type A #
        if not mix_mode:
            A_bool = tissue_types == cell_A
        else:
            col_A = [
                col for i, col in enumerate(cell_type_props.columns) if cell_A in col
            ][0]
            A_bool = cell_type_props.loc[:, col_A].values > cell_prop_cutoff

        A_gene1_bool = np.logical_and(A_bool, gene1_bool)
        A_gene1_sig_bool = np.logical_and(A_gene1_bool, sig_bool)
        A_gene1_sig_indices = np.where(A_gene1_sig_bool)[0]

        for j, cell_B in enumerate(all_set):  # receiver if trans_dir else transmitter
            cellA_cellB_counts = len(
                edge_core(
                    cell_data,
                    j,
                    neighbourhood_bcs,
                    neighbourhood_indices,
                    spot_indices=A_gene1_sig_indices,
                    neigh_bool=gene2_bool,
                    cutoff=cell_prop_cutoff,
                )
            )
            int_matrix[i, j] = cellA_cellB_counts

    return int_matrix if trans_dir else int_matrix.transpose()


def get_interaction_pvals(
    int_matrix,
    n_perms,
    cell_data,
    neighbourhood_bcs,
    neighbourhood_indices,
    all_set,
    sig_bool,
    L_bool,
    R_bool,
    cell_prop_cutoff,
):
    """ Perturbs the cell labels to get background count frequency to estimate \
        p-values.
    """

    # Counting how many times permutation of spots cell data creates interaction
    # counts greater than that observed, in order to calculate p-values.
    greater_counts = np.zeros(int_matrix.shape).astype(int)
    indices = np.array([i for i in range(cell_data.shape[0])])
    # If dealing with discrete data, no need to randomise columns indendently #
    discrete = np.all(np.logical_or(cell_data == 0, cell_data == 1))
    for i in range(n_perms):
        # Permuting the cell data by swapping between spots for each column #
        if not discrete:
            perm_data = cell_data.copy()
            for j in range(cell_data.shape[1]):
                rand_indices = np.random.choice(indices, cell_data.shape[0], False)
                perm_data[:, j] = cell_data[rand_indices, j]
        else:
            rand_indices = np.random.choice(indices, cell_data.shape[0], False)
            perm_data = cell_data[rand_indices, :]

        # Calculating interactions for permuted labels #
        perm_matrix = get_interaction_matrix(
            perm_data,
            neighbourhood_bcs,
            neighbourhood_indices,
            all_set,
            sig_bool,
            L_bool,
            R_bool,
            cell_prop_cutoff,
        ).astype(int)
        perm_greater = (perm_matrix >= int_matrix).astype(int)
        greater_counts += perm_greater

    # Calculating the pvalues #
    int_pvals = greater_counts / n_perms
    return int_pvals


@njit
def get_interaction_matrix(
    cell_data,
    neighbourhood_bcs,
    neighbourhood_indices,
    all_set,
    sig_bool,
    L_bool,
    R_bool,
    cell_prop_cutoff,
):
    """Gets the interaction count matrix."""
    # Now counting the interactions under 3 situations:
    # 1) sig spot with ligand, only neighbours with receptor relevant
    # 2) sig spot with receptor, only neighbours with ligand relevant
    # NOTE, A<->B is double counted, but on different side of matrix.
    # (if bidirectional interaction between two spots, counts as two seperate interactions).
    LR_edges = get_interactions(
        cell_data,
        neighbourhood_bcs,
        neighbourhood_indices,
        all_set,
        sig_bool,
        L_bool,
        R_bool,
        cell_prop_cutoff=cell_prop_cutoff,
        # sig ligand->receptor mode
    )
    RL_edges = get_interactions(
        cell_data,
        neighbourhood_bcs,
        neighbourhood_indices,
        all_set,
        sig_bool,
        R_bool,
        L_bool,
        cell_prop_cutoff=cell_prop_cutoff,
        # sig receptor->ligand mode
    )

    # Counting the number of unique interacting edges
    # between different cell type via indicate LR
    int_matrix = np.zeros((all_set.shape[0], all_set.shape[0]))
    edge_i = 0
    for i in range(all_set.shape[0]):
        for j in range(all_set.shape[0]):
            RL_Atrans_Bedges = LR_edges[edge_i]
            LR_Atrans_Bedges = RL_edges[edge_i]
            edge_i += 1
            max_len = max([len(RL_Atrans_Bedges), len(LR_Atrans_Bedges)])
            if max_len == 0:  # Nothing to count #
                continue

            edge_starts = List()
            edge_ends = List()
            for k in range(max_len):
                if k < len(RL_Atrans_Bedges):
                    edge_starts.append(RL_Atrans_Bedges[k][0])
                    edge_ends.append(RL_Atrans_Bedges[k][1])
                if k < len(LR_Atrans_Bedges):
                    edge_starts.append(LR_Atrans_Bedges[k][0])
                    edge_ends.append(LR_Atrans_Bedges[k][1])
            Atrans_Bedges = List()
            Atrans_Bedges.append((edge_starts[0], edge_ends[0]))  # for typing
            add_unique_edges(Atrans_Bedges, edge_starts, edge_ends)
            # Atrans_Bedges = np.unique(RL_Atrans_Bedges + LR_Atrans_Bedges)
            int_matrix[i, j] = len(Atrans_Bedges) - 1  # since added edge for type

    return int_matrix


@njit
def get_interactions(
    cell_data,
    neighbourhood_bcs,
    neighbourhood_indices,
    all_set,
    sig_bool,
    gene1_bool,
    gene2_bool,
    cell_prop_cutoff=None,
):
    """ Gets spot edges between cell types where the first cell type fits \
        criteria of gene1_bool, & second second cell type of gene2_bool.
    """

    # Creating list of lists to store edges for respective cell types #
    interaction_edges = List()

    # Now retrieving the interaction edges #
    for i in range(all_set.shape[0]):

        # Determining which spots have cell type A #
        A_bool_2 = cell_data[:, i] > cell_prop_cutoff
        A_gene1_bool = np.logical_and(A_bool_2, gene1_bool)

        A_gene1_sig_bool = np.logical_and(A_gene1_bool, sig_bool)
        n_true = A_gene1_sig_bool.sum()
        A_gene1_sig_indices = np.zeros((1, n_true), dtype=np.int_)[
            0, :
        ]  # np.where(A_gene1_sig_bool)[0]
        index = 0
        for k in range(A_gene1_sig_bool.shape[0]):
            if A_gene1_sig_bool[k]:
                A_gene1_sig_indices[index] = k
                index += 1

        for j in range(all_set.shape[0]):
            edge_list = edge_core(
                cell_data,
                j,
                neighbourhood_bcs,
                neighbourhood_indices,
                spot_indices=A_gene1_sig_indices,
                neigh_bool=gene2_bool,
                cutoff=cell_prop_cutoff,
            )

            interaction_edges.append(edge_list)

    return interaction_edges


def create_grids(adata: AnnData, num_row: int, num_col: int, radius: int = 1):
    """Generate screening grids across the tissue sample
    Parameters
    ----------
    adata: AnnData          The data object to generate grids on
    num_row: int            Number of rows
    num_col: int            Number of columns
    radius: int             Radius to determine neighbours (default: 1, nearest)

    Returns
    -------
    grids                 The individual grids defined by left and upper side
    width                   Width of grids
    height                  Height of grids
    """

    from itertools import chain

    coor = adata.obs[["imagerow", "imagecol"]]
    max_x = max(coor["imagecol"])
    min_x = min(coor["imagecol"])
    max_y = max(coor["imagerow"])
    min_y = min(coor["imagerow"])
    width = (max_x - min_x) / num_col
    height = (max_y - min_y) / num_row
    grids, neighbours = [], []
    # generate grids from top to bottom and left to right
    for n in range(num_row * num_col):
        neighbour = []
        x = min_x + n // num_row * width  # left side
        y = min_y + n % num_row * height  # upper side
        grids.append([x, y])

        # get neighbouring grids
        row = n % num_row
        col = n // num_row
        a = np.arange(num_row * num_col).reshape(num_col, num_row).T
        nb_matrix = [
            [
                a[i][j] if 0 <= i < a.shape[0] and 0 <= j < a.shape[1] else -1
                for j in range(col - radius, col + 1 + radius)
            ]
            for i in range(row - radius, row + 1 + radius)
        ]
        for item in nb_matrix:
            neighbour = chain(neighbour, item)
        neighbour = list(set(list(neighbour)))
        neighbours.append(
            [
                grid
                for grid in neighbour
                if not (grid == n and radius > 0) and grid != -1
            ]
        )

    return grids, width, height, neighbours


def count_grid(
    adata: AnnData,
    num_row: int = 30,
    num_col: int = 30,
    use_label: str = None,
    use_het: str = "cci_het_grid",
    radius: int = 1,
    verbose: bool = True,
) -> AnnData:
    """Count the cell type densities
    Parameters
    ----------
    adata: AnnData          The data object including the cell types to count
    num_row: int            Number of grids on height
    num_col: int            Number of grids on width
    use_label:         The cell type results to use in counting
    use_het:                The stoarge place for result
    radius: int             Distance to determine the neighbour grids (default: 1=nearest), radius=0 means within grid

    Returns
    -------
    adata: AnnData          With the counts of specified clusters in each grid of the tissue stored as adata.uns['het']
    """

    coor = adata.obs[["imagerow", "imagecol"]]
    grids, width, height, neighbours = create_grids(adata, num_row, num_col, radius)
    counts = pd.DataFrame(0, range(len(grids)), ["CT"])
    for n, grid in enumerate(grids):
        spots = coor[
            (coor["imagecol"] > grid[0])
            & (coor["imagecol"] < grid[0] + width)
            & (coor["imagerow"] < grid[1])
            & (coor["imagerow"] > grid[1] - height)
        ]
        counts.loc[n] = (adata.obsm[use_label].loc[spots.index] > 0.2).sum().sum()
    adata.obsm[use_het] = (counts / counts.max())["CT"]

    if verbose:
        print(
            "Counts for cluster (cell type) diversity stored into data.uns['"
            + use_het
            + "']"
        )

    return adata



from .plotting.gene_plot import gene_plot
from .plotting.gene_plot import gene_plot_interactive
from .plotting.feat_plot import feat_plot
from .plotting.cluster_plot import cluster_plot
from .plotting.cluster_plot import cluster_plot_interactive
from .plotting.subcluster_plot import subcluster_plot
from .plotting.non_spatial_plot import non_spatial_plot
from .plotting.deconvolution_plot import deconvolution_plot
from .plotting.stack_3d_plot import stack_3d_plot
from .plotting import trajectory
from .plotting.QC_plot import QC_plot
from .plotting.cci_plot import het_plot

# from .plotting.cci_plot import het_plot_interactive
from .plotting.cci_plot import lr_plot_interactive, spatialcci_plot_interactive
from .plotting.cci_plot import grid_plot
from .plotting.cci_plot import lr_diagnostics, lr_n_spots, lr_summary, lr_go
from .plotting.cci_plot import lr_plot, lr_result_plot
from .plotting.cci_plot import (
    ccinet_plot,
    cci_map,
    lr_cci_map,
    lr_chord_plot,
    cci_check,
)
from .plotting.mask_plot import plot_mask



"""
Helper functions for porting R code into python/stLearn.
"""

import os

ro = None
pandas2ri = None
localconverter = None


def rpy2_setup(r_path):
    """Sets up rpy2."""
    os.environ["R_HOME"] = r_path

    import rpy2.robjects as robjects_
    from rpy2.robjects import pandas2ri as pandas2ri_
    from rpy2.robjects.conversion import localconverter as localconverter_

    global ro, pandas2ri, localconverter
    ro = robjects_
    pandas2ri = pandas2ri_
    localconverter = localconverter_



from numpy import linalg as la


def lambda_dist(A1, A2, k=None, p=2, kind="laplacian"):
    """The function is migrated from NetComp package. The lambda distance between graphs, which is defined as
        d(G1,G2) = norm(L_1 - L_2)
    where L_1 is a vector of the top k eigenvalues of the appropriate matrix
    associated with G1, and L2 is defined similarly.
    Parameters
    ----------
    A1, A2 : NumPy matrices
        Adjacency matrices of graphs to be compared
    k : Integer
        The number of eigenvalues to be compared
    p : non-zero Float
        The p-norm is used to compare the resulting vector of eigenvalues.

    kind : String , in {'laplacian','laplacian_norm','adjacency'}
        The matrix for which eigenvalues will be calculated.
    Returns
    -------
    dist : float
        The distance between the two graphs
    Notes
    -----
    The norm can be any p-norm; by default we use p=2. If p<0 is used, the
    result is not a mathematical norm, but may still be interesting and/or
    useful.
    If k is provided, then we use the k SMALLEST eigenvalues for the Laplacian
    distances, and we use the k LARGEST eigenvalues for the adjacency
    distance. This is because the corresponding order flips, as L = D-A.
    References
    ----------
    See Also
    --------
    netcomp.linalg._eigs
    normalized_laplacian_eigs
    """
    # ensure valid k
    n1, n2 = [A.shape[0] for A in [A1, A2]]
    N = min(n1, n2)  # minimum size between the two graphs
    if k is None or k > N:
        k = N

    # form matrices
    L1, L2 = [laplacian_matrix(A) for A in [A1, A2]]
    # get eigenvalues, ignore eigenvectors
    evals1, evals2 = [_eigs(L)[0] for L in [L1, L2]]

    dist = la.norm(evals1[:k] - evals2[:k], ord=p)
    return dist


def resistance_distance(
    A1, A2, p=2, renormalized=False, attributed=False, check_connected=True, beta=1
):
    """Compare two graphs using resistance distance (possibly renormalized).
    Parameters
    ----------
    A1, A2 : NumPy Matrices
        Adjacency matrices of graphs to be compared.
    p : float
        Function returns the p-norm of the flattened matrices.
    renormalized : Boolean, optional (default = False)
        If true, then renormalized resistance distance is computed.
    attributed : Boolean, optional (default=False)
        If true, then the resistance distance PER NODE is returned.
    check_connected : Boolean, optional (default=True)
        If false, then no check on connectivity is performed. See Notes of
        resistance_matrix for more information.
    beta : float, optional (default=1)
        A parameter used in the calculation of the renormalized resistance
        matrix. If using regular resistance, this is irrelevant.
    Returns
    -------
    dist : float of numpy array
        The RR distance between the two graphs. If attributed is True, then
        vector distance per node is returned.
    Notes
    -----
    The distance is calculated by assuming the nodes are in correspondence, and
    any nodes not present are treated as isolated by renormalized resistance.
    References
    ----------
    See Also
    --------
    resistance_matrix
    """
    # Calculate resistance matricies and compare
    if renormalized:
        # pad smaller adj. mat. so they're the same size
        n1, n2 = [A.shape[0] for A in [A1, A2]]
        N = max(n1, n2)
        A1, A2 = [_pad(A, N) for A in [A1, A2]]
        R1, R2 = [renormalized_res_mat(A, beta=beta) for A in [A1, A2]]
    else:
        R1, R2 = [
            resistance_matrix(A, check_connected=check_connected) for A in [A1, A2]
        ]
    try:
        distance_vector = np.sum((R1 - R2) ** p, axis=1)
    except ValueError:
        raise InputError(
            "Input matrices are different sizes. Please use "
            "renormalized resistance distance."
        )
    if attributed:
        return distance_vector ** (1 / p)
    else:
        return np.sum(distance_vector) ** (1 / p)


# **********
# Eigenstuff
# **********
# Functions for calculating eigenstuff of graphs.


from scipy import sparse as sps
import numpy as np
from scipy.sparse import linalg as spla
from numpy import linalg as la

from scipy.sparse import issparse

######################
## Helper Functions ##
######################


def _eigs(M, which="SR", k=None):
    """Helper function for getting eigenstuff.
    Parameters
    ----------
    M : matrix, numpy or scipy sparse
        The matrix for which we hope to get eigenstuff.
    which : string in {'SR','LR'}
        If 'SR', get eigenvalues with smallest real part. If 'LR', get largest.
    k : int
        Number of eigenvalues to return
    Returns
    -------
    evals, evecs : numpy arrays
        Eigenvalues and eigenvectors of matrix M, sorted in ascending or
        descending order, depending on 'which'.
    See Also
    --------
    numpy.linalg.eig
    scipy.sparse.eigs
    """
    n, _ = M.shape
    if k is None:
        k = n
    if which not in ["LR", "SR"]:
        raise ValueError("which must be either 'LR' or 'SR'.")
    M = M.astype(float)
    if issparse(M) and k < n - 1:
        evals, evecs = spla.eigs(M, k=k, which=which)
    else:
        try:
            M = M.todense()
        except:
            pass
        evals, evecs = la.eig(M)
        # sort dem eigenvalues
        inds = np.argsort(evals)
        if which == "LR":
            inds = inds[::-1]
        else:
            pass
        inds = inds[:k]
        evals = evals[inds]
        evecs = np.array(evecs[:, inds])
    return np.real(evals), np.real(evecs)


#####################
##  Get Eigenstuff ##
#####################


def normalized_laplacian_eig(A, k=None):
    """Return the eigenstuff of the normalized Laplacian matrix of graph
    associated with adjacency matrix A.
    Calculates via eigenvalues if
    K = D^(-1/2) A D^(-1/2)
    where `A` is the adjacency matrix and `D` is the diagonal matrix of
    node degrees. Since L = I - K, the eigenvalues and vectors of L can
    be easily recovered.
    Parameters
    ----------
    A : NumPy matrix
        Adjacency matrix of a graph
    k : int, 0 < k < A.shape[0]-1
        The number of eigenvalues to grab.
    Returns
    -------
    lap_evals : NumPy array
       Eigenvalues of L
    evecs : NumPy matrix
       Columns are the eigenvectors of L
    Notes
    -----
    This way of calculating the eigenvalues of the normalized graph laplacian is
    more numerically stable than simply forming the matrix L = I - K and doing
    numpy.linalg.eig on the result. This is because the eigenvalues of L are
    close to zero, whereas the eigenvalues of K are close to 1.
    References
    ----------
    See Also
    --------
    nx.laplacian_matrix
    nx.normalized_laplacian_matrix
    """
    n, m = A.shape
    ##
    ## TODO: implement checks on the adjacency matrix
    ##
    degs = _flat(A.sum(axis=1))
    # the below will break if
    inv_root_degs = [d ** (-1 / 2) if d > _eps else 0 for d in degs]
    inv_rootD = sps.spdiags(inv_root_degs, [0], n, n, format="csr")
    # build normalized diffusion matrix
    K = inv_rootD * A * inv_rootD
    evals, evecs = _eigs(K, k=k, which="LR")
    lap_evals = 1 - evals
    return np.real(lap_evals), np.real(evecs)


#     """
# ********
# Matrices
# ********
# Matrices associated with graphs. Also contains linear algebraic helper functions.
# """


from scipy import sparse as sps
from scipy.sparse import issparse
import numpy as np

_eps = 10 ** (-10)  # a small parameter

######################
## Helper Functions ##
######################


def _flat(D):
    """Flatten column or row matrices, as well as arrays."""
    if issparse(D):
        raise ValueError("Cannot flatten sparse matrix.")
    d_flat = np.array(D).flatten()
    return d_flat


def _pad(A, N):
    """Pad A so A.shape is (N,N)"""
    n, _ = A.shape
    if n >= N:
        return A
    else:
        if issparse(A):
            # thrown if we try to np.concatenate sparse matrices
            side = sps.csr_matrix((n, N - n))
            bottom = sps.csr_matrix((N - n, N))
            A_pad = sps.hstack([A, side])
            A_pad = sps.vstack([A_pad, bottom])
        else:
            side = np.zeros((n, N - n))
            bottom = np.zeros((N - n, N))
            A_pad = np.concatenate([A, side], axis=1)
            A_pad = np.concatenate([A_pad, bottom])
        return A_pad


########################
## Matrices of Graphs ##
########################


def degree_matrix(A):
    """Diagonal degree matrix of graph with adjacency matrix A
    Parameters
    ----------
    A : matrix
        Adjacency matrix
    Returns
    -------
    D : SciPy sparse matrix
        Diagonal matrix of degrees.
    """
    n, m = A.shape
    degs = _flat(A.sum(axis=1))
    D = sps.spdiags(degs, [0], n, n, format="csr")
    return D


def laplacian_matrix(A, normalized=False):
    """Diagonal degree matrix of graph with adjacency matrix A
    Parameters
    ----------
    A : matrix
        Adjacency matrix
    normalized : Bool, optional (default=False)
        If true, then normalized laplacian is returned.
    Returns
    -------
    L : SciPy sparse matrix
        Combinatorial laplacian matrix.
    """
    n, m = A.shape
    D = degree_matrix(A)
    L = D - A
    if normalized:
        degs = _flat(A.sum(axis=1))
        rootD = sps.spdiags(np.power(degs, -1 / 2), [0], n, n, format="csr")
        L = rootD * L * rootD
    return L


# """
# **********
# Exceptions
# **********
# Custom exceptions for NetComp.
# """


class UndefinedException(Exception):
    """Raised when matrix to be returned is undefined"""


# """
# **********
# Resistance
# **********
# Resistance matrix. Renormalized version, as well as conductance and commute matrices.
# """

import networkx as nx
from numpy import linalg as la
from scipy import linalg as spla
import numpy as np
from scipy.sparse import issparse

# from netcomp.linalg.matrices import laplacian_matrix
# from netcomp.exception import UndefinedException


def resistance_matrix(A, check_connected=True):
    """Return the resistance matrix of G.
    Parameters
    ----------
    A : NumPy matrix or SciPy sparse matrix
        Adjacency matrix of a graph.
    check_connected : Boolean, optional (default=True)
        If false, then the resistance matrix will be computed even for
        disconnected matrices. See Notes.
    Returns
    -------
    R : NumPy matrix
       Matrix of pairwise resistances between nodes.
    Notes
    -----
    Uses formula for resistance matrix R in terms of Moore-Penrose of
    pseudoinverse (non-normalized) graph Laplacian. See e.g. Theorem 2.1 in [1].
    This formula can be computed even for disconnected graphs, although the
    interpretation in this case is unclear. Thus, the usage of
    check_connected=False is recommended only to reduce computation time in a
    scenario in which the user is confident the graph in question is, in fact,
    connected.
    Since we do not expect the pseudoinverse of the laplacian to be sparse, we
    convert L to dense form before running np.linalg.pinv(). The returned
    resistance matrix is dense.
    See Also
    --------
    nx.laplacian_matrix
    References
    ----------
    .. [1] W. Ellens, et al. (2011)
       Effective graph resistance.
       Linear Algebra and its Applications, 435 (2011)
    """
    n, m = A.shape
    # check if graph is connected
    if check_connected:
        if issparse(A):
            G = nx.from_scipy_sparse_array(A)
        else:
            G = nx.from_numpy_array(A)
        if not nx.is_connected(G):
            raise UndefinedException(
                "Graph is not connected. " "Resistance matrix is undefined."
            )
    L = laplacian_matrix(A)
    try:
        L = L.todense()
    except:
        pass
    M = la.pinv(L)
    # calculate R in terms of M
    d = np.reshape(np.diag(M), (n, 1))
    ones = np.ones((n, 1))
    R = np.dot(d, ones.T) + np.dot(ones, d.T) - M - M.T
    return R


def commute_matrix(A):
    """Return the commute matrix of the graph associated with adj. matrix A.
    Parameters
    ----------
    A : NumPy matrix or SciPy sparse matrix
        Adjacency matrix of a graph.
    Returns
    -------
    C : NumPy matrix
       Matrix of pairwise resistances between nodes.
    Notes
    -----
    Uses formula for commute time matrix in terms of resistance matrix,
    C = R*2*|E|
    where |E| is the number of edges in G. See e.g. Theorem 2.8 in [1].
    See Also
    --------
    laplacian_matrix
    resistance_matrix
    References
    ----------
    .. [1] W. Ellens, et al. (2011)
       Effective graph resistance.
       Linear Algebra and its Applications, 435 (2011)
    """
    R = resistance_matrix(A)
    E = A.sum() / 2  # number of edges in graph
    C = 2 * E * R
    return C


def renormalized_res_mat(A, beta=1):
    """Return the renormalized resistance matrix of graph associated with A.
    To renormalize a resistance R, we apply the function
    R' = R / (R + beta)
    In this way, the renormalized resistance of nodes in disconnected components
    is 1. The parameter beta determines the penalty for disconnection. If we set
    beta to be approximately the maximum resistance found in the network, then
    the penalty for disconnection is at least 1/2.
    Parameters
    ----------
    A : NumPy matrix or SciPy sparse matrix
        Adjacency matrix of a graph.
    nodelist : list, optional
       The rows and columns are ordered according to the nodes in nodelist. If
       nodelist is None, then the ordering is produced by G.nodes().
    weight : string or None, optional (default='weight')
       The edge data key used to compute each value in the matrix.
       If None, then each edge has weight 1.
    beta : float, optional
       Scaling parameter in renormalization. Must be greater than or equal to
       1. Determines how heavily disconnection is penalized.
    Returns
    -------
    R :  NumPy array
       Matrix of pairwise renormalized resistances between nodes.
    Notes
    -----
    This function converts to a NetworkX graph, as it uses the algorithms
    therein for identifying connected components.
    See Also
    --------
    resistance_matrix
    """
    if issparse(A):
        G = nx.from_scipy_sparse_array(A)
    else:
        G = nx.from_numpy_array(A)
    n = len(G)
    subgraphR = []
    for subgraph in nx.connected_component_subgraphs(G):
        a_sub = nx.adjacency_matrix(subgraph)
        r_sub = resistance_matrix(a_sub)
        subgraphR.append(r_sub)
    R = spla.block_diag(*subgraphR)
    # now, resort R so that it matches the original node list
    component_order = []
    for component in nx.connected_components(G):
        component_order += list(component)
    component_order = list(np.argsort(component_order))
    R = R[component_order, :]
    R = R[:, component_order]
    renorm = np.vectorize(lambda r: r / (r + beta))
    R = renorm(R)
    # set resistance for different components to 1
    R[R == 0] = 1
    R = R - np.eye(n)  # don't want diagonal to be 1
    return R


def conductance_matrix(A):
    """Return the conductance matrix of G.
    The conductance matrix of G is the element-wise inverse of the resistance
    matrix. The diagonal is set to 0, although it is formally infinite. Nodes in
    disconnected components have 0 conductance.
    Parameters
    ----------
    G : graph
       A NetworkX graph
    nodelist : list, optional
       The rows and columns are ordered according to the nodes in nodelist. If
       nodelist is None, then the ordering is produced by G.nodes().
    weight : string or None, optional (default='weight')
       The edge data key used to compute each value in the matrix.
       If None, then each edge has weight 1.
    Returns
    -------
    C :  NumPy array
       Matrix of pairwise conductances between nodes.
    See Also
    --------
    resistance_matrix
    renormalized_res_mat
    """
    if issparse(A):
        G = nx.from_scipy_sparse_array(A)
    else:
        G = nx.from_numpy_array(A)
    subgraphC = []
    for subgraph in nx.connected_component_subgraphs(G):
        a_sub = nx.adjacency_matrix(subgraph)
        r_sub = resistance_matrix(a_sub)
        m = len(subgraph)
        # add one to diagonal, invert, remove one from diagonal:
        c_sub = 1 / (r_sub + np.eye(m)) - np.eye(m)
        subgraphC.append(c_sub)
    C = spla.block_diag(*subgraphC)
    # resort C so that it matches the original node list
    component_order = []
    for component in nx.connected_components(G):
        component_order += list(component)
    component_order = list(np.argsort(component_order))
    C = C[component_order, :]
    C = C[:, component_order]
    return C


########################
## CytoTrace wrapper  ##
########################

from typing import (
    Any,
    Dict,
    List,
    Tuple,
    Union,
    TypeVar,
    Hashable,
    Iterable,
    Optional,
    Sequence,
)
import numpy as np
import pandas as pd
from pandas import Series
from scipy.stats import norm
from numpy.linalg import norm as d_norm
from scipy.sparse import eye as speye
from scipy.sparse import diags, issparse, spmatrix, csr_matrix, isspmatrix_csr
from sklearn.cluster import KMeans
from pandas.api.types import infer_dtype, is_categorical_dtype
from scipy.sparse.linalg import norm as sparse_norm


def _mat_mat_corr_sparse(
    X: csr_matrix,
    Y: np.ndarray,
) -> np.ndarray:
    """\
    This function is borrow from cellrank
    """
    n = X.shape[1]

    X_bar = np.reshape(np.array(X.mean(axis=1)), (-1, 1))
    X_std = np.reshape(
        np.sqrt(np.array(X.power(2).mean(axis=1)) - (X_bar**2)), (-1, 1)
    )

    y_bar = np.reshape(np.mean(Y, axis=0), (1, -1))
    y_std = np.reshape(np.std(Y, axis=0), (1, -1))

    with np.warnings.catch_warnings():
        np.warnings.filterwarnings(
            "ignore", r"invalid value encountered in true_divide"
        )
        return (X @ Y - (n * X_bar * y_bar)) / ((n - 1) * X_std * y_std)


def _correlation_test_helper(
    X: Union[np.ndarray, spmatrix],
    Y: np.ndarray,
    n_perms: Optional[int] = None,
    seed: Optional[int] = None,
    confidence_level: float = 0.95,
    **kwargs,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    This function is borrow from cellrank.
    Compute the correlation between rows in matrix ``X`` columns of matrix ``Y``.
    Parameters
    ----------
    X
        Array or matrix of `(M, N)` elements.
    Y
        Array of `(N, K)` elements.
    method
        Method for p-value calculation.
    n_perms
        Number of permutations if ``method='perm_test'``.
    seed
        Random seed if ``method='perm_test'``.
    confidence_level
        Confidence level for the confidence interval calculation. Must be in `[0, 1]`.
    kwargs
        Keyword arguments for :func:`cellrank.ul._parallelize.parallelize`.
    Returns
    -------
        Correlations, p-values, corrected p-values, lower and upper bound of 95% confidence interval.
        Each array if of shape ``(n_genes, n_lineages)``.
    """

    def perm_test_extractor(
        res: Sequence[Tuple[np.ndarray, np.ndarray]]
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        pvals, corr_bs = zip(*res)
        pvals = np.sum(pvals, axis=0) / float(n_perms)

        corr_bs = np.concatenate(corr_bs, axis=0)
        corr_ci_low, corr_ci_high = np.quantile(corr_bs, q=ql, axis=0), np.quantile(
            corr_bs, q=qh, axis=0
        )

        return pvals, corr_ci_low, corr_ci_high

    if not (0 <= confidence_level <= 1):
        raise ValueError(
            f"Expected `confidence_level` to be in interval `[0, 1]`, found `{confidence_level}`."
        )

    n = X.shape[1]  # genes x cells
    ql = 1 - confidence_level - (1 - confidence_level) / 2.0
    qh = confidence_level + (1 - confidence_level) / 2.0

    if issparse(X) and not isspmatrix_csr(X):
        X = csr_matrix(X)

    corr = _mat_mat_corr_sparse(X, Y) if issparse(X) else _mat_mat_corr_dense(X, Y)

    # see: https://en.wikipedia.org/wiki/Pearson_correlation_coefficient#Using_the_Fisher_transformation
    mean, se = np.arctanh(corr), 1.0 / np.sqrt(n - 3)
    z_score = (np.arctanh(corr) - np.arctanh(0)) * np.sqrt(n - 3)

    z = norm.ppf(qh)
    corr_ci_low = np.tanh(mean - z * se)
    corr_ci_high = np.tanh(mean + z * se)
    pvals = 2 * norm.cdf(-np.abs(z_score))

    return corr, pvals, corr_ci_low, corr_ci_high


def _mat_mat_corr_dense(X: np.ndarray, Y: np.ndarray) -> np.ndarray:
    n = X.shape[1]

    X_bar = np.reshape(np_mean(X, axis=1), (-1, 1))
    X_std = np.reshape(np_std(X, axis=1), (-1, 1))

    y_bar = np.reshape(np_mean(Y, axis=0), (1, -1))
    y_std = np.reshape(np_std(Y, axis=0), (1, -1))

    with np.warnings.catch_warnings():
        np.warnings.filterwarnings(
            "ignore", r"invalid value encountered in true_divide"
        )
        return (X @ Y - (n * X_bar * y_bar)) / ((n - 1) * X_std * y_std)


def _np_apply_along_axis(func1d, axis: int, arr: np.ndarray) -> np.ndarray:
    """
    Apply a reduction function over a given axis.
    Parameters
    ----------
    func1d
        Reduction function that operates only on 1 dimension.
    axis
        Axis over which to apply the reduction.
    arr
        The array to be reduced.
    Returns
    -------
    :class:`numpy.ndarray`
        The reduced array.
    """

    assert arr.ndim == 2
    assert axis in [0, 1]

    if axis == 0:
        result = np.empty(arr.shape[1])
        for i in range(len(result)):
            result[i] = func1d(arr[:, i])
        return result

    result = np.empty(arr.shape[0])
    for i in range(len(result)):
        result[i] = func1d(arr[i, :])

    return result


def np_mean(array: np.ndarray, axis: int) -> np.ndarray:  # noqa
    return _np_apply_along_axis(np.mean, axis, array)


def np_std(array: np.ndarray, axis: int) -> np.ndarray:  # noqa
    return _np_apply_along_axis(np.std, axis, array)



#!/usr/bin/env python

"""Tests for `stlearn` package."""


import unittest

import stlearn as st
import scanpy as sc
from .utils import read_test_data

global adata
adata = read_test_data()


class TestSME(unittest.TestCase):
    """Tests for `stlearn` package."""

    def test_SME(self):
        sc.pp.pca(adata)
        st.pp.tiling(adata, "./tiling")
        st.pp.extract_feature(adata)
        import shutil

        shutil.rmtree("./tiling")
        data_SME = adata.copy()
        # apply stSME to normalise log transformed data
        st.spatial.SME.SME_normalize(data_SME, use_data="raw")



from anndata import AnnData
from typing import Optional, Union
import numpy as np
import pandas as pd
import networkx as nx
from scipy.spatial.distance import cdist
from stlearn.utils import _read_graph
from sklearn.metrics import pairwise_distances


def global_level(
    adata: AnnData,
    use_label: str = "louvain",
    use_rep: str = "X_pca",
    n_dims: int = 40,
    list_clusters: list = [],
    return_graph: bool = False,
    w: float = None,
    verbose: bool = True,
    copy: bool = False,
) -> Optional[AnnData]:

    """\
    Perform global sptial trajectory inference.

    Parameters
    ----------
    adata
        Annotated data matrix.
    list_clusters
        Setup a list of cluster to perform pseudo-space-time
    use_label
        Use label result of cluster method.
    return_graph
        Return PTS graph
    w
        Pseudo-spatio-temporal distance weight (balance between spatial effect and DPT)
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Anndata
    """

    assert w <= 1, "w should be in range 0 to 1"
    # Get global graph
    G = _read_graph(adata, "global_graph")
    # Convert to directed graph
    H = G.to_directed()

    cat_inds = adata.uns[use_label + "_index_dict"]
    inds_cat = {v: k for (k, v) in cat_inds.items()}

    # Query cluster
    if type(list_clusters[0]) == str:
        list_clusters = [cat_inds[label] for label in list_clusters]
    query_nodes = list_clusters

    query_nodes = ordering_nodes(query_nodes, use_label, adata)
    if verbose:
        print(
            "Start to construct the trajectory: "
            + " -> ".join(np.array(query_nodes).astype(str))
        )

    query_dict = {}
    order_dict = {}

    for i in query_nodes:
        order = 0
        for j in adata.obs[adata.obs[use_label] == str(inds_cat[i])][
            "sub_cluster_labels"
        ].unique():
            query_dict[int(j)] = int(i)
            order_dict[int(j)] = int(order)

            order += 1
    dm_list = []
    sdm_list = []
    order_big_dict = {}
    edge_list = []

    for i, j in enumerate(query_nodes):
        order_big_dict[j] = int(i)
        if i == len(query_nodes) - 1:
            break
        for j in adata.uns["split_node"][query_nodes[i]]:
            for k in adata.uns["split_node"][query_nodes[i + 1]]:
                edge_list.append((int(j), int(k)))

        # Calculate DPT distance matrix
        dm_list.append(
            ge_distance_matrix(
                adata,
                inds_cat[query_nodes[i]],
                inds_cat[query_nodes[i + 1]],
                use_label=use_label,
                use_rep=use_rep,
                n_dims=n_dims,
            )
        )
        # Calculate Spatial distance matrix
        sdm_list.append(
            spatial_distance_matrix(
                adata,
                inds_cat[query_nodes[i]],
                inds_cat[query_nodes[i + 1]],
                use_label=use_label,
            )
        )

    # Get centroid dictionary
    centroid_dict = adata.uns["centroid_dict"]
    centroid_dict = {int(key): centroid_dict[key] for key in centroid_dict}

    H_sub = H.edge_subgraph(edge_list)
    if not nx.is_connected(H_sub.to_undirected()):
        raise ValueError(
            "The chosen clusters are not available to construct the spatial trajectory! Please choose other path."
        )
    H_sub = nx.DiGraph(H_sub)
    prepare_root = []
    for node in adata.uns["split_node"][query_nodes[0]]:
        H_sub.add_edge(9999, int(node))
        prepare_root.append(centroid_dict[int(node)])

    prepare_root = np.array(prepare_root)
    centroide = (
        sum(prepare_root[:, 0]) / len(prepare_root[:, 0]),
        sum(prepare_root[:, 1]) / len(prepare_root[:, 1]),
    )

    # Get centroid dictionary
    centroid_dict = adata.uns["centroid_dict"]
    centroid_dict = {int(key): centroid_dict[key] for key in centroid_dict}

    H_sub = H.edge_subgraph(edge_list)
    H_sub = nx.DiGraph(H_sub)

    prepare_root = []
    for node in adata.uns["split_node"][query_nodes[0]]:
        H_sub.add_edge(9999, int(node))
        prepare_root.append(centroid_dict[int(node)])

    prepare_root = np.array(prepare_root)
    centroide = (
        sum(prepare_root[:, 0]) / len(prepare_root[:, 0]),
        sum(prepare_root[:, 1]) / len(prepare_root[:, 1]),
    )
    centroid_dict[9999] = np.array(centroide)

    labels = nx.get_edge_attributes(H_sub, "weight")

    for edge, _ in labels.items():

        dm = dm_list[order_big_dict[query_dict[edge[0]]]]
        sdm = sdm_list[order_big_dict[query_dict[edge[0]]]]

        weight = dm[order_dict[edge[0]], order_dict[edge[1]]] * w + sdm[
            order_dict[edge[0]], order_dict[edge[1]]
        ] * (1 - w)
        H_sub[edge[0]][edge[1]]["weight"] = weight
    # tmp = H_sub

    H_sub = nx.algorithms.tree.minimum_spanning_arborescence(H_sub)
    H_nodes = list(range(len(H_sub.nodes)))

    node_convert = {}
    for pair in zip(list(H_sub.nodes), H_nodes):
        node_convert[pair[1]] = pair[0]

    adata.uns["PTS_graph"] = {}
    adata.uns["PTS_graph"]["graph"] = nx.to_scipy_sparse_array(H_sub)
    adata.uns["PTS_graph"]["node_dict"] = node_convert

    if return_graph:
        return H_sub


########################
## Global level PTS ##
########################


def get_node(node_list, split_node):
    result = np.array([])
    for node in node_list:
        result = np.append(result, np.array(split_node[node]).astype(int))
    return result.astype(int)


def ordering_nodes(node_list, use_label, adata):
    mean_dpt = []
    for node in node_list:
        mean_dpt.append(
            adata.obs[adata.obs[use_label] == str(node)]["dpt_pseudotime"].median()
        )

    return list(np.array(node_list)[np.argsort(mean_dpt)])


# def dpt_distance_matrix(adata, cluster1, cluster2, use_label):
#     tmp = adata.obs[adata.obs[use_label] == str(cluster1)]
#     chosen_adata1 = adata[list(tmp.index)]
#     tmp = adata.obs[adata.obs[use_label] == str(cluster2)]
#     chosen_aadata = adata[list(tmp.index)]

#     sub_dpt1 = []
#     chosen_sub1 = chosen_adata1.obs["sub_cluster_labels"].unique()
#     for i in range(0, len(chosen_sub1)):
#         sub_dpt1.append(
#             chosen_adata1.obs[
#                 chosen_adata1.obs["sub_cluster_labels"] == chosen_sub1[i]
#             ]["dpt_pseudotime"].median()
#         )

#     sub_dpt2 = []
#     chosen_sub2 = chosen_aadata.obs["sub_cluster_labels"].unique()
#     for i in range(0, len(chosen_sub2)):
#         sub_dpt2.append(
#             chosen_aadata.obs[
#                 chosen_aadata.obs["sub_cluster_labels"] == chosen_sub2[i]
#             ]["dpt_pseudotime"].median()
#         )

#     dm = cdist(
#         np.array(sub_dpt1).reshape(-1, 1),
#         np.array(sub_dpt2).reshape(-1, 1),
#         lambda u, v: v - u,
#     )
#     from sklearn.preprocessing import MinMaxScaler
#     scaler = MinMaxScaler()
#     scale_dm = scaler.fit_transform(dm)
#     # scale_dm = (dm + (-np.min(dm))) / np.max(dm)
#     return scale_dm


def spatial_distance_matrix(adata, cluster1, cluster2, use_label):

    tmp = adata.obs[adata.obs[use_label] == str(cluster1)]
    chosen_adata1 = adata[list(tmp.index)]
    tmp = adata.obs[adata.obs[use_label] == str(cluster2)]
    chosen_aadata = adata[list(tmp.index)]

    centroid_dict = adata.uns["centroid_dict"]
    centroid_dict = {int(key): centroid_dict[key] for key in centroid_dict}

    sub_coord1 = []
    chosen_sub1 = chosen_adata1.obs["sub_cluster_labels"].unique()
    for i in range(0, len(chosen_sub1)):
        sub_coord1.append(centroid_dict[int(chosen_sub1[i])])

    sub_coord2 = []
    chosen_sub2 = chosen_aadata.obs["sub_cluster_labels"].unique()
    for i in range(0, len(chosen_sub2)):
        sub_coord2.append(centroid_dict[int(chosen_sub2[i])])

    sdm = cdist(np.array(sub_coord1), np.array(sub_coord2), "euclidean")

    from sklearn.preprocessing import MinMaxScaler

    # scaler = MinMaxScaler()
    # scale_sdm = scaler.fit_transform(sdm)
    scale_sdm = sdm / np.max(sdm)

    return scale_sdm


def ge_distance_matrix(adata, cluster1, cluster2, use_label, use_rep, n_dims):

    tmp = adata.obs[adata.obs[use_label] == str(cluster1)]
    chosen_adata1 = adata[list(tmp.index)]
    tmp = adata.obs[adata.obs[use_label] == str(cluster2)]
    chosen_aadata = adata[list(tmp.index)]

    centroid_dict = adata.uns["centroid_dict"]
    centroid_dict = {int(key): centroid_dict[key] for key in centroid_dict}

    sub_coord1 = []
    chosen_sub1 = chosen_adata1.obs["sub_cluster_labels"].unique()
    for i in chosen_sub1:
        sub_coord1.append(
            np.array(
                chosen_adata1[chosen_adata1.obs["sub_cluster_labels"].isin([i])].obsm[
                    use_rep
                ][:, :n_dims]
            )
        )

    sub_coord2 = []
    chosen_sub2 = chosen_aadata.obs["sub_cluster_labels"].unique()
    for i in chosen_sub2:
        sub_coord2.append(
            np.array(
                chosen_aadata[chosen_aadata.obs["sub_cluster_labels"].isin([i])].obsm[
                    use_rep
                ][:, :n_dims]
            )
        )

    results = []
    for i in range(0, len(sub_coord1)):
        for j in range(0, len(sub_coord2)):
            results.append(cdist(sub_coord1[i], sub_coord2[j], "cosine").mean())
    results = np.array(results).reshape(len(sub_coord1), len(sub_coord2))

    from sklearn.preprocessing import MinMaxScaler

    # scaler = MinMaxScaler()
    # scale_sdm = scaler.fit_transform(results)
    scale_sdm = results / np.max(results)

    return scale_sdm


# def _density_normalize(other: Union[np.ndarray, spmatrix]
#     ) -> Union[np.ndarray, spmatrix]:
#         """
#         Density normalization by the underlying KNN graph.
#         Parameters
#         ----------
#         other:
#             Matrix to normalize.
#         Returns
#         -------
#         :class:`np.ndarray` or :class:`scipy.sparse.spmatrix`
#             Density normalized transition matrix.
#         """

#         q = np.asarray(_conn.sum(axis=0))

#         if not issparse(other):
#             Q = np.diag(1.0 / q)
#         else:
#             Q = spdiags(1.0 / q, 0, other.shape[0], other.shape[0])

#         return Q @ other @ Q


# def compute_transition_matrix(
#     adata: AnnData,
#     ):

# def compute_spatial_trans_probs(
#     adata: AnnData,
#     ):



from sklearn.metrics import pairwise_distances
from typing import Optional, Union
from anndata import AnnData
import numpy as np
from ..._compat import Literal
from tqdm import tqdm

_PLATFORM = Literal["Visium", "Old_ST"]
_WEIGHTING_MATRIX = Literal[
    "weights_matrix_all",
    "weights_matrix_pd_gd",
    "weights_matrix_pd_md",
    "weights_matrix_gd_md",
    "gene_expression_correlation",
    "physical_distance",
    "morphological_distance",
]


def calculate_weight_matrix(
    adata: AnnData,
    adata_imputed: Union[AnnData, None] = None,
    pseudo_spots: bool = False,
    platform: _PLATFORM = "Visium",
) -> Optional[AnnData]:
    from sklearn.linear_model import LinearRegression
    import math

    if platform == "Visium":
        img_row = adata.obs["imagerow"]
        img_col = adata.obs["imagecol"]
        array_row = adata.obs["array_row"]
        array_col = adata.obs["array_col"]
        rate = 3
    elif platform == "Old_ST":
        img_row = adata.obs["imagerow"]
        img_col = adata.obs["imagecol"]
        array_row = adata.obs_names.map(lambda x: x.split("x")[1])
        array_col = adata.obs_names.map(lambda x: x.split("x")[0])
        rate = 1.5
    else:
        raise ValueError(
            f"""\
                {platform!r} does not support.
                """
        )

    reg_row = LinearRegression().fit(array_row.values.reshape(-1, 1), img_row)

    reg_col = LinearRegression().fit(array_col.values.reshape(-1, 1), img_col)

    if pseudo_spots and adata_imputed:
        pd = pairwise_distances(
            adata_imputed.obs[["imagecol", "imagerow"]],
            adata.obs[["imagecol", "imagerow"]],
            metric="euclidean",
        )
        unit = math.sqrt(reg_row.coef_**2 + reg_col.coef_**2)
        pd_norm = np.where(pd >= unit, 0, 1)

        md = 1 - pairwise_distances(
            adata_imputed.obsm["X_morphology"],
            adata.obsm["X_morphology"],
            metric="cosine",
        )
        md[md < 0] = 0

        adata_imputed.uns["physical_distance"] = pd_norm
        adata_imputed.uns["morphological_distance"] = md

        adata_imputed.uns["weights_matrix_all"] = (
            adata_imputed.uns["physical_distance"]
            * adata_imputed.uns["morphological_distance"]
        )

    else:
        pd = pairwise_distances(adata.obs[["imagecol", "imagerow"]], metric="euclidean")
        unit = math.sqrt(reg_row.coef_**2 + reg_col.coef_**2)
        pd_norm = np.where(pd >= rate * unit, 0, 1)

        md = 1 - pairwise_distances(adata.obsm["X_morphology"], metric="cosine")
        md[md < 0] = 0

        gd = 1 - pairwise_distances(adata.obsm["X_pca"], metric="correlation")
        adata.uns["gene_expression_correlation"] = gd
        adata.uns["physical_distance"] = pd_norm
        adata.uns["morphological_distance"] = md

        adata.uns["weights_matrix_all"] = (
            adata.uns["physical_distance"]
            * adata.uns["morphological_distance"]
            * adata.uns["gene_expression_correlation"]
        )
        adata.uns["weights_matrix_pd_gd"] = (
            adata.uns["physical_distance"] * adata.uns["gene_expression_correlation"]
        )
        adata.uns["weights_matrix_pd_md"] = (
            adata.uns["physical_distance"] * adata.uns["morphological_distance"]
        )
        adata.uns["weights_matrix_gd_md"] = (
            adata.uns["gene_expression_correlation"]
            * adata.uns["morphological_distance"]
        )


def impute_neighbour(
    adata: AnnData,
    count_embed: Union[np.ndarray, None] = None,
    weights: _WEIGHTING_MATRIX = "weights_matrix_all",
    copy: bool = False,
) -> Optional[AnnData]:
    coor = adata.obs[["imagecol", "imagerow"]]

    weights_matrix = adata.uns[weights]

    lag_coor = []

    weights_list = []

    with tqdm(
        total=len(adata),
        desc="Adjusting data",
        bar_format="{l_bar}{bar} [ time left: {remaining} ]",
    ) as pbar:
        for i in range(len(coor)):

            main_weights = weights_matrix[i]

            if weights == "physical_distance":
                current_neighbour = main_weights.argsort()[-6:]
            else:
                current_neighbour = main_weights.argsort()[-3:]

            surrounding_count = count_embed[current_neighbour]
            surrounding_weights = main_weights[current_neighbour]
            if surrounding_weights.sum() > 0:
                surrounding_weights_scaled = (
                    surrounding_weights / surrounding_weights.sum()
                )
                weights_list.append(surrounding_weights_scaled)

                surrounding_count_adjusted = np.multiply(
                    surrounding_weights_scaled.reshape(-1, 1), surrounding_count
                )
                surrounding_count_final = np.sum(surrounding_count_adjusted, axis=0)

            else:
                surrounding_count_final = np.zeros(count_embed.shape[1])
                weights_list.append(np.zeros(len(current_neighbour)))
            lag_coor.append(surrounding_count_final)
            pbar.update(1)

    imputed_data = np.array(lag_coor)
    key_added = "imputed_data"
    adata.obsm[key_added] = imputed_data

    adata.obsm["top_weights"] = np.array(weights_list)

    return adata if copy else None



#!/usr/bin/env python

"""Tests for `stlearn` package."""


import unittest

import stlearn as st
import scanpy as sc
from .utils import read_test_data
import numpy as np

global adata
adata = read_test_data()


class TestPSTS(unittest.TestCase):
    """Tests for `stlearn` package."""

    def test_PSTS(self):
        sc.pp.pca(adata)
        sc.pp.neighbors(adata)
        sc.tl.leiden(adata, resolution=0.6)
        sc.tl.louvain(adata)

        adata.uns["iroot"] = np.flatnonzero(adata.obs["leiden"] == "0")[0]
        st.spatial.trajectory.pseudotime(
            adata, eps=100, use_rep="X_pca", use_sme=False, use_label="leiden"
        )
        st.spatial.trajectory.pseudotimespace_global(
            adata, use_label="leiden", list_clusters=[0, 1]
        )
        st.spatial.trajectory.detect_transition_markers_clades(
            adata, clade=0, use_raw_count=False, cutoff_spearman=0.3
        )



import numpy as np
import pandas as pd
import scipy as sc
from numba import njit, prange
from numba.typed import List
import scipy.spatial as spatial
from anndata import AnnData
from .het import create_grids


def lr(
    adata: AnnData,
    use_lr: str = "cci_lr",
    distance: float = None,
    verbose: bool = True,
    neighbours: list = None,
    fast: bool = True,
) -> AnnData:

    """Calculate the proportion of known ligand-receptor co-expression among the neighbouring spots or within spots
    Parameters
    ----------
    adata: AnnData          The data object to scan
    use_lr: str             object to keep the result (default: adata.uns['cci_lr'])
    distance: float         Distance to determine the neighbours (default: closest), distance=0 means within spot
    neighbours: list        List of the neighbours for each spot, if None then computed. Useful for speeding up function.
    fast: bool              Whether to use the fast implimentation or not.

    Returns
    -------
    adata: AnnData          The data object including the results
    """

    # automatically calculate distance if not given, won't overwrite distance=0 which is within-spot
    distance = calc_distance(adata, distance)

    # # expand the LR pairs list by swapping ligand-receptor positions
    lr_pairs = adata.uns["lr"].copy()
    spot_lr1 = get_spot_lrs(adata, lr_pairs=lr_pairs, lr_order=True)
    spot_lr2 = get_spot_lrs(adata, lr_pairs=lr_pairs, lr_order=False)
    if verbose:
        print("Altogether " + str(spot_lr1.shape[1]) + " valid L-R pairs")

    # get neighbour spots for each spot according to the specified distance
    if type(neighbours) == type(None):
        neighbours = calc_neighbours(adata, distance, index=fast)

    # Calculating the scores, can have either the fast or the pandas version #
    if fast:
        adata.obsm[use_lr] = lr_core(spot_lr1.values, spot_lr2.values, neighbours, 0)
    else:
        adata.obsm[use_lr] = lr_pandas(spot_lr1, spot_lr2, neighbours)

    if verbose:
        print(
            "L-R interactions with neighbours are counted and stored into adata.obsm['"
            + use_lr
            + "']"
        )

    # return adata


def calc_distance(adata: AnnData, distance: float):
    """Automatically calculate distance if not given, won't overwrite \
        distance=0 which is within-spot.
    Parameters
    ----------
    adata: AnnData          The data object to scan
    distance: float         Distance to determine the neighbours (default: closest), distance=0 means within spot

    Returns
    -------
    distance: float         The automatically calcualted distance (or inputted distance)
    """
    if not distance and distance != 0:
        # for arranged-spots
        scalefactors = next(iter(adata.uns["spatial"].values()))["scalefactors"]
        library_id = list(adata.uns["spatial"].keys())[0]
        distance = (
            scalefactors["spot_diameter_fullres"]
            * scalefactors[
                "tissue_" + adata.uns["spatial"][library_id]["use_quality"] + "_scalef"
            ]
            * 2
        )
    return distance


def get_lrs_scores(
    adata: AnnData,
    lrs: np.array,
    neighbours: np.array,
    het_vals: np.array,
    min_expr: float,
    filter_pairs: bool = True,
    spot_indices: np.array = None,
):
    """Gets the scores for the indicated set of LR pairs & the heterogeneity values.
    Parameters
    ----------
    adata: AnnData   See run() doc-string.
    lrs: np.array    See run() doc-string.
    neighbours: np.array    Array of arrays with indices specifying neighbours of each spot.
    het_vals: np.array      Cell heterogeneity counts per spot.
    min_expr: float         Minimum gene expression of either L or R for spot to be considered to have reasonable score.
    filter_pairs: bool      Whether to filter to valid pairs or not.
    spot_indices: np.array  Array of integers speci
    Returns
    -------
    lrs: np.array   lr pairs from the database in format ['L1_R1', 'LN_RN']
    """
    if type(spot_indices) == type(None):
        spot_indices = np.array(list(range(len(adata))), dtype=np.int_)

    spot_lr1s = get_spot_lrs(
        adata, lr_pairs=lrs, lr_order=True, filter_pairs=filter_pairs
    )
    spot_lr2s = get_spot_lrs(
        adata, lr_pairs=lrs, lr_order=False, filter_pairs=filter_pairs
    )
    if filter_pairs:
        lrs = np.array(
            [
                "_".join(spot_lr1s.columns.values[i : i + 2])
                for i in range(0, spot_lr1s.shape[1], 2)
            ]
        )

    # Calculating the lr_scores across spots for the inputted lrs #
    lr_scores = get_scores(
        spot_lr1s.values, spot_lr2s.values, neighbours, het_vals, min_expr, spot_indices
    )

    if filter_pairs:
        return lr_scores, lrs
    else:
        return lr_scores


def get_spot_lrs(
    adata: AnnData,
    lr_pairs: list,
    lr_order: bool,
    filter_pairs: bool = True,
):
    """
    Parameters
    ----------
    adata: AnnData         The adata object to scan
    lr_pairs: list         List of the lr pairs (e.g. ['L1_R1', 'L2_R2',...]
    lr_order: bool         Forward version of the spot lr pairs (L1_R1), False indicates reverse (R1_L1)
    filter_pairs: bool     Whether to filter the pairs or not (check if present before subsetting).

    Returns
    -------
    spot_lrs: pd.DataFrame   Spots*GeneOrder, in format l1, r1, ... ln, rn if lr_order True, else r1, l1, ... rn, ln
    """
    df = adata.to_df()
    pairs_rev = [f'{pair.split("_")[1]}_{pair.split("_")[0]}' for pair in lr_pairs]
    pairs_wRev = []
    for i in range(len(lr_pairs)):
        pairs_wRev.extend([lr_pairs[i], pairs_rev[i]])

    if filter_pairs:
        pairs_wRev = [
            lr
            for i, lr in enumerate(pairs_wRev)
            if lr.split("_")[0] in df.columns and lr.split("_")[1] in df.columns
        ]

    lr_cols = [pair.split("_")[int(lr_order == False)] for pair in pairs_wRev]
    spot_lrs = df[lr_cols]
    return spot_lrs


def calc_neighbours(
    adata: AnnData,
    distance: float = None,
    index: bool = True,
    verbose: bool = True,
) -> List:
    """Calculate the proportion of known ligand-receptor co-expression among the neighbouring spots or within spots
    Parameters
    ----------
    adata: AnnData          The data object to scan
    distance: float         Distance to determine the neighbours (default: closest), distance=0 means within spot
    index: bool             Indicates whether to return neighbours as indices to other spots or names of other spots.

    Returns
    -------
    neighbours: numba.typed.List          List of np.array's indicating neighbours by indices for each spot.
    """
    if verbose:
        print("Calculating neighbours...")
    # get neighbour spots for each spot according to the specified distance
    coor = adata.obs[["imagerow", "imagecol"]]
    point_tree = spatial.cKDTree(coor)
    neighbours = []
    for i, spot in enumerate(adata.obs_names):
        if distance == 0:
            neighbours.append(np.array([i if index else spot]))
        else:
            n_index = point_tree.query_ball_point(
                np.array(
                    [adata.obs["imagerow"].loc[spot], adata.obs["imagecol"].loc[spot]]
                ),
                distance,
            )
            if index:
                n_index = np.array(n_index, dtype=np.int_)
                neighbours.append(n_index[n_index != i])
            else:
                n_spots = adata.obs_names[n_index]
                neighbours.append(n_spots[n_spots != spot])

    typed_neighs = List()
    [typed_neighs.append(neigh) for neigh in neighbours]

    n_neighs = np.array([len(neigh) for neigh in neighbours])
    if verbose:
        print(
            f"{len(np.where(n_neighs==0)[0])} spots with no neighbours, "
            f"{int(np.median(n_neighs))} median spot neighbours."
        )

    if np.all(n_neighs == 0):
        raise Exception(
            "All spots have no neighbours at current distance,"
            " set distance to higher value, or distance=0 for "
            "within-spot mode."
        )
    return typed_neighs


@njit
def lr_core(
    spot_lr1: np.ndarray,
    spot_lr2: np.ndarray,
    neighbours: List,
    min_expr: float,
    spot_indices: np.array,
) -> np.ndarray:
    """Calculate the lr scores for each spot.
    Parameters
    ----------
    spot_lr1: np.ndarray          Spots*Ligands
    spot_lr2: np.ndarray          Spots*Receptors
    neighbours: numba.typed.List          List of np.array's indicating neighbours by indices for each spot.
    min_expr: float               Minimum expression for gene to be considered expressed.
    Returns
    -------
    lr_scores: numpy.ndarray   Cells*LR-scores.
    """
    # Calculating mean of lr2 expressions from neighbours of each spot
    nb_lr2 = np.zeros((len(spot_indices), spot_lr2.shape[1]), np.float64)
    for i in range(len(spot_indices)):
        spot_i = spot_indices[i]
        nb_expr = spot_lr2[neighbours[spot_i], :]
        if nb_expr.shape[0] != 0:  # Accounting for no neighbours
            nb_expr_mean = nb_expr.sum(axis=0) / nb_expr.shape[0]
        else:
            nb_expr_mean = nb_expr.sum(axis=0)
        nb_lr2[i, :] = nb_expr_mean

    scores = (
        spot_lr1[spot_indices, :] * (nb_lr2 > min_expr)
        + (spot_lr1[spot_indices, :] > min_expr) * nb_lr2
    )
    spot_lr = scores.sum(axis=1)
    return spot_lr / 2


def lr_pandas(
    spot_lr1: np.ndarray,
    spot_lr2: np.ndarray,
    neighbours: list,
) -> np.ndarray:
    """Calculate the lr scores for each spot.
    Parameters
    ----------
    spot_lr1: pd.DataFrame          Cells*Ligands
    spot_lr2: pd.DataFrame          Cells*Receptors
    neighbours: list       List of neighbours by indices for each spot.
    Returns
    -------
    lr_scores: numpy.ndarray   Cells*LR-scores.
    """
    # function to calculate mean of lr2 expression between neighbours or within spot (distance==0) for each spot
    def mean_lr2(x):
        # get lr2 expressions from the neighbour(s)
        n_spots = neighbours[spot_lr2.index.tolist().index(x.name)]
        nbs = spot_lr2.loc[n_spots, :]
        if nbs.shape[0] > 0:  # if neighbour exists
            return nbs.sum() / nbs.shape[0]
        else:
            return 0

    # mean of lr2 expressions from neighbours of each spot
    nb_lr2 = spot_lr2.apply(mean_lr2, axis=1)

    # check whether neighbours exist
    try:
        nb_lr2.shape[1]
    except:
        raise ValueError("No neighbours found within given distance.")

    # keep value of nb_lr2 only when lr1 is also expressed on the spots
    spot_lr = pd.DataFrame(
        spot_lr1.values * (nb_lr2.values > 0) + (spot_lr1.values > 0) * nb_lr2.values,
    ).sum(axis=1)
    return spot_lr.values / 2


@njit(parallel=True)
def get_scores(
    spot_lr1s: np.ndarray,
    spot_lr2s: np.ndarray,
    neighbours: List,
    het_vals: np.array,
    min_expr: float,
    spot_indices: np.array,
) -> np.array:
    """Calculates the scores.
    Parameters
    ----------
    spot_lr1s: np.ndarray   Spots*GeneOrder1, in format l1, r1, ... ln, rn
    spot_lr2s: np.ndarray   Spots*GeneOrder2, in format r1, l1, ... rn, ln
    het_vals:  np.ndarray   Spots*Het counts
    neighbours: numba.typed.List          List of np.array's indicating neighbours by indices for each spot.
    min_expr: float               Minimum expression for gene to be considered expressed.
    Returns
    -------
    spot_scores: np.ndarray   Spots*LR pair of the LR scores per spot.
    """
    spot_scores = np.zeros((len(spot_indices), spot_lr1s.shape[1] // 2), np.float64)
    for i in prange(0, spot_lr1s.shape[1] // 2):
        i_ = i * 2  # equivalent to range(0, spot_lr1s.shape[1], 2)
        spot_lr1, spot_lr2 = spot_lr1s[:, i_ : (i_ + 2)], spot_lr2s[:, i_ : (i_ + 2)]
        lr_scores = lr_core(spot_lr1, spot_lr2, neighbours, min_expr, spot_indices)
        # The merge scores #
        lr_scores = np.multiply(het_vals[spot_indices], lr_scores)
        spot_scores[:, i] = lr_scores
    return spot_scores


def lr_grid(
    adata: AnnData,
    num_row: int = 10,
    num_col: int = 10,
    use_lr: str = "cci_lr_grid",
    radius: int = 1,
    verbose: bool = True,
) -> AnnData:

    """Calculate the proportion of known ligand-receptor co-expression among the neighbouring grids or within each grid
    Parameters
    ----------
    adata: AnnData          The data object to scan
    num_row: int            Number of grids on height
    num_col: int            Number of grids on width
    use_lr: str             object to keep the result (default: adata.uns['cci_lr'])
    radius: int             Distance to determine the neighbour grids (default: 1=nearest), radius=0 means within grid

    Returns
    -------
    adata: AnnData          The data object with the cci_lr grid result updated
    """

    # prepare data as pd.dataframe
    df = adata.to_df()
    if not isinstance(df, pd.DataFrame):
        if sc.sparse.issparse(df):
            df = pd.DataFrame(
                df.toarray(), index=adata.obs_names, columns=adata.var_names
            )
        else:
            df = pd.DataFrame(df, index=adata.obs_names, columns=adata.var_names)

    # generate grids and prepare grid level dataframe
    grids, width, height, neighbours = create_grids(adata, num_row, num_col, radius)
    df_grid = pd.DataFrame(0, index=range(len(grids)), columns=df.columns)

    # transform spot level into grid level
    coor = adata.obs[["imagerow", "imagecol"]]
    for n, grid in enumerate(grids):
        spots = coor[
            (coor["imagecol"] > grid[0])
            & (coor["imagecol"] < grid[0] + width)
            & (coor["imagerow"] < grid[1])
            & (coor["imagerow"] > grid[1] - height)
        ]
        df_grid.loc[n] = df.loc[spots.index].sum()

    # expand the LR pairs list by swapping ligand-receptor positions
    lr_pairs = adata.uns["lr"].copy()
    lr_pairs += [item.split("_")[1] + "_" + item.split("_")[0] for item in lr_pairs]

    # filter out those LR pairs not expressing in the dataset
    lr1 = [item.split("_")[0] for item in lr_pairs]
    lr2 = [item.split("_")[1] for item in lr_pairs]
    avail = [
        i for i, x in enumerate(lr1) if lr1[i] in df.columns and lr2[i] in df.columns
    ]
    grid_lr1 = df_grid[[lr1[i] for i in avail]]
    grid_lr2 = df_grid[[lr2[i] for i in avail]]
    if verbose:
        print("Altogether " + str(len(avail)) + " valid L-R pairs")

    # function to calculate mean of lr2 expression between neighbours or within spot (distance==0) for each spot
    def mean_lr2(x):
        # get the neighbour(s)' lr2 expressions
        nbs = grid_lr2.loc[neighbours[df_grid.index.tolist().index(x.name)], :]
        if nbs.shape[0] > 0:  # if neighbour exists
            return nbs.sum() / nbs.shape[0]
        else:
            return 0

    # mean of lr2 expression among neighbour grids
    nb_lr2 = grid_lr2.apply(mean_lr2, axis=1)

    # keep value of nb_lr2 only when lr1 is also expressed on the grids
    grid_lr = pd.DataFrame(
        grid_lr1.values * (nb_lr2.values > 0) + (grid_lr1.values > 0) * nb_lr2.values,
        index=df_grid.index,
        columns=[lr_pairs[i] for i in avail],
    ).sum(axis=1)
    adata.obsm[use_lr] = grid_lr.values / 2

    if verbose:
        print(
            "L-R interactions with neighbours are counted and stored into adata.uns['"
            + use_lr
            + "']"
        )

    return adata



from anndata import AnnData
from typing import Optional, Union
from .weight_optimization import weight_optimizing_global, weight_optimizing_local
from .global_level import global_level
from .local_level import local_level


def pseudotimespace_global(
    adata: AnnData,
    use_label: str = "louvain",
    use_rep: str = "X_pca",
    n_dims: int = 40,
    list_clusters: list = [],
    model: str = "spatial",
    step=0.01,
    k=10,
) -> Optional[AnnData]:

    """\
    Perform pseudo-time-space analysis with global level.

    Parameters
    ----------
    adata
        Annotated data matrix.
    use_label
        Use label result of cluster method.
    list_clusters
        List of cluster used to reconstruct spatial trajectory.
    w
        Weighting factor to balance between spatial data and gene expression
    step
        Step for screeing weighting factor
    k
        The number of eigenvalues to be compared
    Returns
    -------
    Anndata
    """

    if model == "mixed":

        w = weight_optimizing_global(
            adata, use_label=use_label, list_clusters=list_clusters, step=step, k=k
        )
    elif model == "spatial":
        w = 0
    elif model == "gene_expression":
        w = 1
    else:
        raise ValidationError(
            "Please choose the right model! Available models: 'mixed', 'spatial' and 'gene_expression' "
        )

    global_level(
        adata,
        use_label=use_label,
        list_clusters=list_clusters,
        w=w,
        use_rep=use_rep,
        n_dims=n_dims,
    )


def pseudotimespace_local(
    adata: AnnData,
    use_label: str = "louvain",
    cluster: list = [],
    w: float = None,
) -> Optional[AnnData]:

    """\
    Perform pseudo-time-space analysis with local level.

    Parameters
    ----------
    adata
        Annotated data matrix.
    use_label
        Use label result of cluster method.
    cluster
        Cluster used to reconstruct intraregional spatial trajectory.
    w
        Weighting factor to balance between spatial data and gene expression
    Returns
    -------
    Anndata
    """

    if w is None:
        w = weight_optimizing_local(adata, use_label=use_label, cluster=cluster)

    local_level(adata, use_label=use_label, cluster=cluster, w=w)



import numpy as np
import pandas as pd
import io
from PIL import Image
import matplotlib
from anndata import AnnData
import networkx as nx

from typing import Optional, Union, Mapping  # Special
from typing import Sequence, Iterable  # ABCs
from typing import Tuple  # Classes

from textwrap import dedent

from enum import Enum


class Empty(Enum):
    token = 0


_empty = Empty.token

from matplotlib import rcParams, ticker, gridspec, axes
from matplotlib.axes import Axes
from abc import ABC


class _AxesSubplot(Axes, axes.SubplotBase, ABC):
    """Intersection between Axes and SubplotBase: Has methods of both"""


def _check_spot_size(
    spatial_data: Optional[Mapping], spot_size: Optional[float]
) -> float:
    """
    Resolve spot_size value.
    This is a required argument for spatial plots.
    """
    if spatial_data is None and spot_size is None:
        raise ValueError(
            "When .uns['spatial'][library_id] does not exist, spot_size must be "
            "provided directly."
        )
    elif spot_size is None:
        return spatial_data["scalefactors"]["spot_diameter_fullres"]
    else:
        return spot_size


def _check_scale_factor(
    spatial_data: Optional[Mapping],
    img_key: Optional[str],
    scale_factor: Optional[float],
) -> float:
    """Resolve scale_factor, defaults to 1."""
    if scale_factor is not None:
        return scale_factor
    elif spatial_data is not None and img_key is not None:
        return spatial_data["scalefactors"][f"tissue_{img_key}_scalef"]
    else:
        return 1.0


def _check_spatial_data(
    uns: Mapping, library_id: Union[Empty, None, str]
) -> Tuple[Optional[str], Optional[Mapping]]:
    """
    Given a mapping, try and extract a library id/ mapping with spatial data.
    Assumes this is `.uns` from how we parse visium data.
    """
    spatial_mapping = uns.get("spatial", {})
    if library_id is _empty:
        if len(spatial_mapping) > 1:
            raise ValueError(
                "Found multiple possible libraries in `.uns['spatial']. Please specify."
                f" Options are:\n\t{list(spatial_mapping.keys())}"
            )
        elif len(spatial_mapping) == 1:
            library_id = list(spatial_mapping.keys())[0]
        else:
            library_id = None
    if library_id is not None:
        spatial_data = spatial_mapping[library_id]
    else:
        spatial_data = None
    return library_id, spatial_data


def _check_img(
    spatial_data: Optional[Mapping],
    img: Optional[np.ndarray],
    img_key: Union[None, str, Empty],
    bw: bool = False,
) -> Tuple[Optional[np.ndarray], Optional[str]]:
    """
    Resolve image for spatial plots.
    """
    if img is None and spatial_data is not None and img_key is _empty:
        img_key = next(
            (k for k in ["hires", "lowres", "fulres"] if k in spatial_data["images"]),
        )  # Throws StopIteration Error if keys not present
    if img is None and spatial_data is not None and img_key is not None:
        img = spatial_data["images"][img_key]
    if bw:
        img = np.dot(img[..., :3], [0.2989, 0.5870, 0.1140])
    return img, img_key


def _check_coords(
    obsm: Optional[Mapping], scale_factor: Optional[float]
) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:

    image_coor = obsm["spatial"] * scale_factor
    imagecol = image_coor[:, 0]
    imagerow = image_coor[:, 1]

    return [imagecol, imagerow]


def _read_graph(adata: AnnData, graph_type: Optional[str]):

    if graph_type == "PTS_graph":
        graph = nx.from_scipy_sparse_array(
            adata.uns[graph_type]["graph"], create_using=nx.DiGraph
        )
    else:
        graph = nx.from_scipy_sparse_array(adata.uns[graph_type]["graph"])
    node_dict = adata.uns[graph_type]["node_dict"]
    node_dict = {int(k): int(v) for k, v in node_dict.items()}

    relabel_graph = nx.relabel_nodes(graph, node_dict)

    return relabel_graph


def _docs_params(**kwds):
    """\
    Docstrings should start with "\" in the first line for proper formatting.
    """

    def dec(obj):
        obj.__orig_doc__ = obj.__doc__
        obj.__doc__ = dedent(obj.__doc__).format_map(kwds)
        return obj

    return dec



import matplotlib.pyplot as plt
from decimal import Decimal
from anndata import AnnData
from typing import Optional, Union


def transition_markers_plot(
    adata: AnnData,
    top_genes: int = 10,
    trajectory: str = None,
    dpi: int = 150,
    output: str = None,
    name: str = None,
) -> Optional[AnnData]:
    """\
    Plot transition marker.

    Parameters
    ----------
    adata
        Annotated data matrix.
    top_genes
        Top genes users want to display in the plot.
    trajectory
        Name of a clade/branch user wants to plot transition markers.
    dpi
        The resolution of the plot.
    output
        The output folder of the plot.
    name
        The filename of the plot.
    Returns
    -------
    Anndata
    """

    if trajectory == None:
        raise ValueError("Please input the trajectory name!")
    if trajectory not in adata.uns:
        raise ValueError("Please input the right trajectory name!")

    pos = (
        adata.uns[trajectory][adata.uns[trajectory]["score"] >= 0]
        .sort_values("score", ascending=False)
        .reset_index(drop=True)[:top_genes]
    )
    neg = (
        adata.uns[trajectory][adata.uns[trajectory]["score"] < 0]
        .sort_values("score")
        .reset_index(drop=True)[:top_genes]
    )

    y = range(top_genes)
    x1 = list(neg["score"])[::-1]
    x2 = list(pos["score"])[::-1]

    pos = pos[::-1]
    neg = neg[::-1]

    if len(x1) < top_genes:
        for i in range(len(x1), top_genes):
            x1.append(0)
    if len(x2) < top_genes:
        for i in range(len(x2), top_genes):
            x2.append(0)

    fig, axes = plt.subplots(ncols=2, sharey=True)
    axes[0].barh(y, x1, align="center", color="#fb687a")
    axes[1].barh(y, x2, align="center", color="#31a2fb")
    fig.subplots_adjust(wspace=0)
    axes[0].spines["left"].set_visible(False)
    axes[0].spines["right"].set_visible(False)
    axes[0].spines["top"].set_visible(False)
    axes[1].spines["left"].set_visible(False)
    axes[1].spines["right"].set_visible(False)
    axes[1].spines["top"].set_visible(False)
    axes[0].get_yaxis().set_ticks([])
    axes[1].get_yaxis().set_ticks([])
    axes[0].tick_params(axis="both", which="both", length=0)
    axes[1].tick_params(axis="both", which="both", length=0)

    for i, x in enumerate([x1, x2]):
        if all(value == 0 for value in x):
            if i == 0:
                axes[0].get_xaxis().set_ticks([])
            if i == 1:
                axes[1].get_xaxis().set_ticks([])

    rects = axes[1].patches
    for i, rect in enumerate(rects):
        try:
            gene_name = list(pos["gene"])[i]
            p_value = "{:.2E}".format(Decimal(str(list(pos["p-value"])[i])))
        except:
            gene_name = ""
            p_value = ""
        alignment = {"horizontalalignment": "left", "verticalalignment": "center"}
        axes[1].text(
            rect.get_x() + rect.get_width() + 0.01,
            rect.get_y() + rect.get_height() / 2.0,
            gene_name,
            **alignment,
            size=6
        )
        axes[1].text(
            rect.get_x() + 0.01,
            rect.get_y() + rect.get_height() / 2.0,
            p_value,
            color="w",
            **alignment,
            size=6
        )

    rects = axes[0].patches
    for i, rect in enumerate(rects):
        try:
            gene_name = list(neg["gene"])[i]
            p_value = "{:.2E}".format(Decimal(str(list(neg["p-value"])[i])))
        except:
            gene_name = ""
            p_value = ""
        alignment = {"horizontalalignment": "right", "verticalalignment": "center"}
        axes[0].text(
            rect.get_x() + rect.get_width() - 0.01,
            rect.get_y() + rect.get_height() / 2.0,
            gene_name,
            **alignment,
            size=6
        )
        axes[0].text(
            rect.get_x() - 0.01,
            rect.get_y() + rect.get_height() / 2.0,
            p_value,
            color="w",
            **alignment,
            size=6
        )

    plt.figtext(0.5, 0.9, trajectory, ha="center", va="center")
    axes[0].set_xlabel("Spearman correlation coefficient")
    axes[0].xaxis.set_label_coords(1, -0.1)

    axes[0].grid(False)
    axes[1].grid(False)

    if name is None:
        name = trajectory

    if output is not None:
        fig.savefig(output + "/" + name, dpi=dpi, bbox_inches="tight", pad_inches=0)

    plt.show()



from .pseudotime_plot import pseudotime_plot
from .local_plot import local_plot
from .tree_plot import tree_plot
from .transition_markers_plot import transition_markers_plot
from .DE_transition_plot import DE_transition_plot
from .check_trajectory import check_trajectory



""" Contains different kinds of form validators.
"""
from wtforms.validators import ValidationError


class CheckNumberRange(object):
    def __init__(self, lower, upper, hint=""):
        self.lower = lower
        self.upper = upper
        self.hint = hint

    def __call__(self, form, field):

        if field.data is not None:
            if not (self.lower <= float(field.data) <= self.upper):
                if self.hint:
                    raise ValidationError(self.hint)
                else:
                    raise ValidationError("Not in correct range")



from matplotlib import pyplot as plt
import numpy as np
from typing import Optional, Union
from anndata import AnnData


def QC_plot(
    adata: AnnData,
    library_id: str = None,
    name: str = None,
    data_alpha: float = 0.8,
    tissue_alpha: float = 1.0,
    cmap: str = "Spectral_r",
    spot_size: tuple = (5, 40),
    show_color_bar: bool = True,
    show_size_legend: bool = True,
    show_axis: bool = False,
    cropped: bool = True,
    margin: int = 100,
    dpi: int = 150,
    output: str = None,
) -> Optional[AnnData]:
    """\
        QC plot for sptial transcriptomics data.

        Parameters
        ----------
        adata
            Annotated data matrix.
        library_id
            Library id stored in AnnData.
        data_alpha
            Opacity of the spot.
        tissue_alpha
            Opacity of the tissue.
        cmap
            Color map to use.
        spot_size
            Size of the spot (min, max).
        show_color_bar
            Show color bar or not.
        show_axis
            Show axis or not.
        show_size_legend
            Show size legend or not.
        name
            Name of the output figure file.
        output
            Save the figure as file or not.
        copy
            Return a copy instead of writing to adata.
        Returns
        -------
        Nothing
        """

    imagecol = adata.obs["imagecol"]
    imagerow = adata.obs["imagerow"]
    from sklearn.preprocessing import MinMaxScaler

    reads_per_spot = adata.to_df().sum(axis=1)
    scaler = MinMaxScaler(feature_range=spot_size)
    reads_per_spot_size = scaler.fit_transform(reads_per_spot.to_numpy().reshape(-1, 1))
    genes_per_spot = adata.to_df().astype(bool).sum(axis=1)

    # plt.rcParams['figure.dpi'] = dpi

    # Option for turning off showing figure
    plt.ioff()

    # Initialize matplotlib
    fig, a = plt.subplots()

    vmin = min(genes_per_spot)
    vmax = max(genes_per_spot)
    # Plot scatter plot based on pixel of spots
    plot = a.scatter(
        adata.obs["imagecol"],
        adata.obs["imagerow"],
        edgecolor="none",
        alpha=data_alpha,
        s=reads_per_spot_size,
        marker="o",
        vmin=vmin,
        vmax=vmax,
        cmap=plt.get_cmap(cmap),
        c=genes_per_spot,
    )

    if show_color_bar:
        from mpl_toolkits.axes_grid1.inset_locator import inset_axes

        axins = inset_axes(
            a,
            width="100%",
            height="100%",
            loc="upper left",
            bbox_to_anchor=(1.0, 0.73, 0.05, 0.35),
            bbox_transform=a.transAxes,
            borderpad=4.3,
        )
        cb = plt.colorbar(plot, cax=axins)
        cb.ax.set_xlabel("Number of Genes", fontsize=10)
        cb.ax.xaxis.set_label_coords(0.98, 1.20)
        cb.outline.set_visible(False)

    if show_size_legend:
        size_min, size_max = spot_size
        markers = [
            size_min,
            size_min + 1 / 3 * (size_max - size_min),
            size_min + 2 / 3 * (size_max - size_min),
            size_max,
        ]
        legend_markers = [plt.scatter([], [], s=i, c="grey") for i in markers]
        labels = [
            str(int(scaler.inverse_transform(np.array(i).reshape(1, 1))))
            for i in markers
        ]
        a.legend(
            handles=legend_markers,
            labels=labels,
            loc="lower left",
            bbox_to_anchor=(1, 0.05),
            scatterpoints=1,
            frameon=False,
            title="Number of Reads",
        )

    if not show_axis:
        a.axis("off")
    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]

    image = adata.uns["spatial"][library_id]["images"][
        adata.uns["spatial"][library_id]["use_quality"]
    ]
    # Overlay the tissue image
    a.imshow(
        image,
        alpha=tissue_alpha,
        zorder=-1,
    )

    if cropped:
        a.set_xlim(imagecol.min() - margin, imagecol.max() + margin)

        a.set_ylim(imagerow.min() - margin, imagerow.max() + margin)

        a.set_ylim(a.get_ylim()[::-1])
        # plt.gca().invert_yaxis()

    # fig.tight_layout()
    if output is not None:
        fig.savefig(output + "/" + name, dpi=dpi, bbox_inches="tight", pad_inches=0)

    plt.show()



from typing import Optional, Union
from anndata import AnnData
from matplotlib import pyplot as plt
from pathlib import Path
import os


def convert_scanpy(
    adata: AnnData,
    use_quality: str = "hires",
) -> Optional[AnnData]:

    adata.var_names_make_unique()

    library_id = list(adata.uns["spatial"].keys())[0]

    if use_quality == "fulres":
        image_coor = adata.obsm["spatial"]
    else:
        scale = adata.uns["spatial"][library_id]["scalefactors"][
            "tissue_" + use_quality + "_scalef"
        ]
        image_coor = adata.obsm["spatial"] * scale

    adata.obs["imagecol"] = image_coor[:, 0]
    adata.obs["imagerow"] = image_coor[:, 1]
    adata.uns["spatial"][library_id]["use_quality"] = use_quality

    return adata



from matplotlib import pyplot as plt
from PIL import Image
import pandas as pd
import matplotlib
import numpy as np

from typing import Optional, Union, Mapping  # Special
from typing import Sequence, Iterable  # ABCs
from typing import Tuple  # Classes

from anndata import AnnData
import warnings

from stlearn.plotting.classes import SubClusterPlot
from stlearn.plotting._docs import doc_spatial_base_plot, doc_subcluster_plot
from stlearn.utils import _AxesSubplot, Axes, _docs_params


@_docs_params(
    spatial_base_plot=doc_spatial_base_plot, subcluster_plot=doc_subcluster_plot
)
def subcluster_plot(
    adata: AnnData,
    # plotting param
    title: Optional["str"] = None,
    figsize: Optional[Tuple[float, float]] = None,
    cmap: Optional[str] = "jet",
    use_label: Optional[str] = None,
    list_clusters: Optional[list] = None,
    ax: Optional[_AxesSubplot] = None,
    show_plot: Optional[bool] = True,
    show_axis: Optional[bool] = False,
    show_image: Optional[bool] = True,
    show_color_bar: Optional[bool] = True,
    crop: Optional[bool] = True,
    margin: Optional[bool] = 100,
    size: Optional[float] = 5,
    image_alpha: Optional[float] = 1.0,
    cell_alpha: Optional[float] = 1.0,
    fname: Optional[str] = None,
    dpi: Optional[int] = 120,
    # subcluster plot param
    cluster: Optional[int] = 0,
    threshold_spots: Optional[int] = 5,
    text_box_size: Optional[float] = 5,
    bbox_to_anchor: Optional[Tuple[float, float]] = (1, 1),
) -> Optional[AnnData]:
    """\
    Allows the visualization of a subclustering results as the discretes values
    of dot points in the Spatial transcriptomics array.

    Parameters
    -------------------------------------
    {spatial_base_plot}
    {subcluster_plot}

    Examples
    -------------------------------------
    >>> import stlearn as st
    >>> adata = st.datasets.example_bcba()
    >>> label = "louvain"
    >>> cluster = 6
    >>> st.pl.cluster_plot(adata, use_label = label, cluster = cluster)

    """

    assert use_label != None, "Please select `use_label` parameter"
    assert (
        use_label in adata.obs.columns
    ), "Please run `stlearn.spatial.cluster.localization` function!"

    SubClusterPlot(
        adata,
        title=title,
        figsize=figsize,
        cmap=cmap,
        use_label=use_label,
        list_clusters=list_clusters,
        ax=ax,
        show_plot=show_plot,
        show_axis=show_axis,
        show_image=show_image,
        show_color_bar=show_color_bar,
        crop=crop,
        margin=margin,
        size=size,
        image_alpha=image_alpha,
        cell_alpha=cell_alpha,
        fname=fname,
        dpi=dpi,
        text_box_size=text_box_size,
        bbox_to_anchor=bbox_to_anchor,
        cluster=cluster,
        threshold_spots=threshold_spots,
    )



from .global_level import global_level
from .local_level import local_level
from .pseudotime import pseudotime
from .weight_optimization import weight_optimizing_global, weight_optimizing_local
from .utils import lambda_dist, resistance_distance
from .pseudotimespace import pseudotimespace_global, pseudotimespace_local
from .detect_transition_markers import (
    detect_transition_markers_clades,
    detect_transition_markers_branches,
)
from .compare_transitions import compare_transitions

from .set_root import set_root



"""Reading and Writing
"""
from pathlib import Path, PurePath
from typing import Optional, Union
from anndata import AnnData
import numpy as np
from PIL import Image
import pandas as pd
import stlearn
from .._compat import Literal
import scanpy
import scipy
import matplotlib.pyplot as plt

_QUALITY = Literal["fulres", "hires", "lowres"]
_background = ["black", "white"]


def Read10X(
    path: Union[str, Path],
    genome: Optional[str] = None,
    count_file: str = "filtered_feature_bc_matrix.h5",
    library_id: str = None,
    load_images: Optional[bool] = True,
    quality: _QUALITY = "hires",
    image_path: Union[str, Path] = None,
) -> AnnData:

    """\
    Read Visium data from 10X (wrap read_visium from scanpy)

    In addition to reading regular 10x output,
    this looks for the `spatial` folder and loads images,
    coordinates and scale factors.
    Based on the `Space Ranger output docs`_.

    .. _Space Ranger output docs: https://support.10xgenomics.com/spatial-gene-expression/software/pipelines/latest/output/overview

    Parameters
    ----------
    path
        Path to directory for visium datafiles.
    genome
        Filter expression to genes within this genome.
    count_file
        Which file in the passed directory to use as the count file. Typically would be one of:
        'filtered_feature_bc_matrix.h5' or 'raw_feature_bc_matrix.h5'.
    library_id
        Identifier for the visium library. Can be modified when concatenating multiple adata objects.
    load_images
        Load image or not.
    quality
        Set quality that convert to stlearn to use. Store in anndata.obs['imagecol' & 'imagerow']
    image_path
        Path to image. Only need when loading full resolution image.


    Returns
    -------
    Annotated data matrix, where observations/cells are named by their
    barcode and variables/genes by gene name. Stores the following information:
    :attr:`~anndata.AnnData.X`
        The data matrix is stored
    :attr:`~anndata.AnnData.obs_names`
        Cell names
    :attr:`~anndata.AnnData.var_names`
        Gene names
    :attr:`~anndata.AnnData.var`\\ `['gene_ids']`
        Gene IDs
    :attr:`~anndata.AnnData.var`\\ `['feature_types']`
        Feature types
    :attr:`~anndata.AnnData.uns`\\ `['spatial']`
        Dict of spaceranger output files with 'library_id' as key
    :attr:`~anndata.AnnData.uns`\\ `['spatial'][library_id]['images']`
        Dict of images (`'fulres'`, `'hires'` and `'lowres'`)
    :attr:`~anndata.AnnData.uns`\\ `['spatial'][library_id]['scalefactors']`
        Scale factors for the spots
    :attr:`~anndata.AnnData.uns`\\ `['spatial'][library_id]['metadata']`
        Files metadata: 'chemistry_description', 'software_version'
    :attr:`~anndata.AnnData.obsm`\\ `['spatial']`
        Spatial spot coordinates, usable as `basis` by :func:`~scanpy.pl.embedding`.
    """

    from scanpy import read_visium

    adata = read_visium(
        path,
        genome=genome,
        count_file=count_file,
        library_id=library_id,
        load_images=load_images,
    )
    adata.var_names_make_unique()

    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]

    if quality == "fulres":
        image_coor = adata.obsm["spatial"]
        img = plt.imread(image_path, 0)
        adata.uns["spatial"][library_id]["images"]["fulres"] = img
    else:
        scale = adata.uns["spatial"][library_id]["scalefactors"][
            "tissue_" + quality + "_scalef"
        ]
        image_coor = adata.obsm["spatial"] * scale

    adata.obs["imagecol"] = image_coor[:, 0]
    adata.obs["imagerow"] = image_coor[:, 1]
    adata.uns["spatial"][library_id]["use_quality"] = quality

    return adata


def ReadOldST(
    count_matrix_file: Union[str, Path] = None,
    spatial_file: Union[str, Path] = None,
    image_file: Union[str, Path] = None,
    library_id: str = "OldST",
    scale: float = 1.0,
    quality: str = "hires",
    spot_diameter_fullres: float = 50,
) -> AnnData:

    """\
    Read Old Spatial Transcriptomics data

    Parameters
    ----------
    count_matrix_file
        Path to count matrix file.
    spatial_file
        Path to spatial location file.
    image_file
        Path to the tissue image file
    library_id
        Identifier for the visium library. Can be modified when concatenating multiple adata objects.
    scale
        Set scale factor.
    quality
        Set quality that convert to stlearn to use. Store in anndata.obs['imagecol' & 'imagerow']
    spot_diameter_fullres
        Diameter of spot in full resolution

    Returns
    -------
    AnnData
    """

    adata = scanpy.read_text(count_matrix_file)
    adata = stlearn.add.parsing(adata, coordinates_file=spatial_file)
    stlearn.add.image(
        adata,
        library_id=library_id,
        quality=quality,
        imgpath=image_file,
        scale=scale,
        spot_diameter_fullres=spot_diameter_fullres,
    )

    return adata


def ReadSlideSeq(
    count_matrix_file: Union[str, Path],
    spatial_file: Union[str, Path],
    library_id: str = None,
    scale: float = None,
    quality: str = "hires",
    spot_diameter_fullres: float = 50,
    background_color: _background = "white",
) -> AnnData:

    """\
    Read Slide-seq data

    Parameters
    ----------
    count_matrix_file
        Path to count matrix file.
    spatial_file
        Path to spatial location file.
    library_id
        Identifier for the visium library. Can be modified when concatenating multiple adata objects.
    scale
        Set scale factor.
    quality
        Set quality that convert to stlearn to use. Store in anndata.obs['imagecol' & 'imagerow']
    spot_diameter_fullres
        Diameter of spot in full resolution
    background_color
        Color of the backgound. Only `black` or `white` is allowed.

    Returns
    -------
    AnnData
    """

    count = pd.read_csv(count_matrix_file)
    meta = pd.read_csv(spatial_file)

    adata = AnnData(count.iloc[:, 1:].set_index("gene").T)

    adata.var["ENSEMBL"] = count["ENSEMBL"].values

    adata.obs["index"] = meta["index"].values

    if scale == None:
        max_coor = np.max(meta[["x", "y"]].values)
        scale = 2000 / max_coor

    adata.obs["imagecol"] = meta["x"].values * scale
    adata.obs["imagerow"] = meta["y"].values * scale

    # Create image
    max_size = np.max([adata.obs["imagecol"].max(), adata.obs["imagerow"].max()])
    max_size = int(max_size + 0.1 * max_size)

    if background_color == "black":
        image = Image.new("RGBA", (max_size, max_size), (0, 0, 0, 0))
    else:
        image = Image.new("RGBA", (max_size, max_size), (255, 255, 255, 255))
    imgarr = np.array(image)

    if library_id is None:
        library_id = "Slide-seq"

    adata.uns["spatial"] = {}
    adata.uns["spatial"][library_id] = {}
    adata.uns["spatial"][library_id]["images"] = {}
    adata.uns["spatial"][library_id]["images"][quality] = imgarr
    adata.uns["spatial"][library_id]["use_quality"] = quality
    adata.uns["spatial"][library_id]["scalefactors"] = {}
    adata.uns["spatial"][library_id]["scalefactors"][
        "tissue_" + quality + "_scalef"
    ] = scale

    adata.uns["spatial"][library_id]["scalefactors"][
        "spot_diameter_fullres"
    ] = spot_diameter_fullres
    adata.obsm["spatial"] = meta[["x", "y"]].values

    return adata


def ReadMERFISH(
    count_matrix_file: Union[str, Path],
    spatial_file: Union[str, Path],
    library_id: str = None,
    scale: float = None,
    quality: str = "hires",
    spot_diameter_fullres: float = 50,
    background_color: _background = "white",
) -> AnnData:

    """\
    Read MERFISH data

    Parameters
    ----------
    count_matrix_file
        Path to count matrix file.
    spatial_file
        Path to spatial location file.
    library_id
        Identifier for the visium library. Can be modified when concatenating multiple adata objects.
    scale
        Set scale factor.
    quality
        Set quality that convert to stlearn to use. Store in anndata.obs['imagecol' & 'imagerow']
    spot_diameter_fullres
        Diameter of spot in full resolution
    background_color
        Color of the backgound. Only `black` or `white` is allowed.

    Returns
    -------
    AnnData
    """

    coordinates = pd.read_excel(spatial_file, index_col=0)
    if coordinates.min().min() < 0:
        coordinates = coordinates + np.abs(coordinates.min().min()) + 100
    from scanpy import read_csv

    counts = read_csv(count_matrix_file).transpose()

    adata_merfish = counts[coordinates.index, :]
    adata_merfish.obsm["spatial"] = coordinates.to_numpy()

    if scale == None:
        max_coor = np.max(adata_merfish.obsm["spatial"])
        scale = 2000 / max_coor

    adata_merfish.obs["imagecol"] = adata_merfish.obsm["spatial"][:, 0] * scale
    adata_merfish.obs["imagerow"] = adata_merfish.obsm["spatial"][:, 1] * scale

    # Create image
    max_size = np.max(
        [adata_merfish.obs["imagecol"].max(), adata_merfish.obs["imagerow"].max()]
    )
    max_size = int(max_size + 0.1 * max_size)
    if background_color == "black":
        image = Image.new("RGB", (max_size, max_size), (0, 0, 0, 0))
    else:
        image = Image.new("RGB", (max_size, max_size), (255, 255, 255, 255))
    imgarr = np.array(image)

    if library_id is None:
        library_id = "MERSEQ"

    adata_merfish.uns["spatial"] = {}
    adata_merfish.uns["spatial"][library_id] = {}
    adata_merfish.uns["spatial"][library_id]["images"] = {}
    adata_merfish.uns["spatial"][library_id]["images"][quality] = imgarr
    adata_merfish.uns["spatial"][library_id]["use_quality"] = quality
    adata_merfish.uns["spatial"][library_id]["scalefactors"] = {}
    adata_merfish.uns["spatial"][library_id]["scalefactors"][
        "tissue_" + quality + "_scalef"
    ] = scale
    adata_merfish.uns["spatial"][library_id]["scalefactors"][
        "spot_diameter_fullres"
    ] = spot_diameter_fullres

    return adata_merfish


def ReadSeqFish(
    count_matrix_file: Union[str, Path],
    spatial_file: Union[str, Path],
    library_id: str = None,
    scale: float = 1.0,
    quality: str = "hires",
    field: int = 0,
    spot_diameter_fullres: float = 50,
    background_color: _background = "white",
) -> AnnData:

    """\
    Read SeqFish data

    Parameters
    ----------
    count_matrix_file
        Path to count matrix file.
    spatial_file
        Path to spatial location file.
    library_id
        Identifier for the visium library. Can be modified when concatenating multiple adata objects.
    scale
        Set scale factor.
    quality
        Set quality that convert to stlearn to use. Store in anndata.obs['imagecol' & 'imagerow']
    field
        Set field of view for SeqFish data
    spot_diameter_fullres
        Diameter of spot in full resolution
    background_color
        Color of the backgound. Only `black` or `white` is allowed.
    Returns
    -------
    AnnData
    """

    count = pd.read_table(count_matrix_file, header=None)
    spatial = pd.read_table(spatial_file, index_col=False)

    count = count.T
    count.columns = count.iloc[0]
    count = count.drop(count.index[0]).reset_index(drop=True)
    count = count[count["Field_of_View"] == field].drop(count.columns[[0, 1]], axis=1)

    spatial = spatial[spatial["Field_of_View"] == field]

    adata = AnnData(count)

    if scale == None:
        max_coor = np.max(spatial[["X", "Y"]])
        scale = 2000 / max_coor

    adata.obs["imagecol"] = spatial["X"].values * scale
    adata.obs["imagerow"] = spatial["Y"].values * scale

    adata.obsm["spatial"] = spatial[["X", "Y"]].values

    # Create image
    max_size = np.max([adata.obs["imagecol"].max(), adata.obs["imagerow"].max()])
    max_size = int(max_size + 0.1 * max_size)

    if background_color == "black":
        image = Image.new("RGBA", (max_size, max_size), (0, 0, 0, 0))
    else:
        image = Image.new("RGBA", (max_size, max_size), (255, 255, 255, 255))
    imgarr = np.array(image)

    if library_id is None:
        library_id = "Slide-seq"

    adata.uns["spatial"] = {}
    adata.uns["spatial"][library_id] = {}
    adata.uns["spatial"][library_id]["images"] = {}
    adata.uns["spatial"][library_id]["images"][quality] = imgarr
    adata.uns["spatial"][library_id]["use_quality"] = quality
    adata.uns["spatial"][library_id]["scalefactors"] = {}
    adata.uns["spatial"][library_id]["scalefactors"][
        "tissue_" + quality + "_scalef"
    ] = scale
    adata.uns["spatial"][library_id]["scalefactors"][
        "spot_diameter_fullres"
    ] = spot_diameter_fullres

    return adata


def ReadXenium(
    feature_cell_matrix_file: Union[str, Path],
    cell_summary_file: Union[str, Path],
    image_path: Optional[Path] = None,
    library_id: str = None,
    scale: float = 1.0,
    quality: str = "hires",
    spot_diameter_fullres: float = 15,
    background_color: _background = "white",
) -> AnnData:

    """\
    Read Xenium data

    Parameters
    ----------
    feature_cell_matrix_file
        Path to feature cell count matrix file. Only support h5ad file now.
    cell_summary_file
        Path to cell summary CSV file.
    image_path
        Path to image. Only need when loading full resolution image.
    library_id
        Identifier for the visium library. Can be modified when concatenating multiple adata objects.
    scale
        Set scale factor.
    quality
        Set quality that convert to stlearn to use. Store in anndata.obs['imagecol' & 'imagerow']
    spot_diameter_fullres
        Diameter of spot in full resolution
    background_color
        Color of the backgound. Only `black` or `white` is allowed.
    Returns
    -------
    AnnData
    """

    metadata = pd.read_csv(cell_summary_file)

    adata = scanpy.read_10x_h5(feature_cell_matrix_file)

    spatial = metadata[["x_centroid", "y_centroid"]]
    spatial.columns = ["imagecol", "imagerow"]

    adata.obsm["spatial"] = spatial.values

    if scale == None:
        max_coor = np.max(adata.obsm["spatial"])
        scale = 2000 / max_coor

    if library_id is None:
        library_id = "Xenium_data"

    adata.obs["imagecol"] = spatial["imagecol"].values * scale
    adata.obs["imagerow"] = spatial["imagerow"].values * scale

    if image_path != None:
        stlearn.add.image(
            adata,
            library_id=library_id,
            quality=quality,
            imgpath=image_path,
            scale=scale,
        )
    else:
        # Create image
        max_size = np.max([adata.obs["imagecol"].max(), adata.obs["imagerow"].max()])
        max_size = int(max_size + 0.1 * max_size)

        if background_color == "black":
            image = Image.new("RGBA", (max_size, max_size), (0, 0, 0, 0))
        else:
            image = Image.new("RGBA", (max_size, max_size), (255, 255, 255, 255))
        imgarr = np.array(image)

        # Create spatial dictionary
        adata.uns["spatial"] = {}
        adata.uns["spatial"][library_id] = {}
        adata.uns["spatial"][library_id]["images"] = {}
        adata.uns["spatial"][library_id]["images"][quality] = imgarr
        adata.uns["spatial"][library_id]["use_quality"] = quality
        adata.uns["spatial"][library_id]["scalefactors"] = {}
        adata.uns["spatial"][library_id]["scalefactors"][
            "tissue_" + quality + "_scalef"
        ] = scale
        adata.uns["spatial"][library_id]["scalefactors"][
            "spot_diameter_fullres"
        ] = spot_diameter_fullres

    return adata


def create_stlearn(
    count: pd.DataFrame,
    spatial: pd.DataFrame,
    library_id: str,
    image_path: Optional[Path] = None,
    scale: float = None,
    quality: str = "hires",
    spot_diameter_fullres: float = 50,
    background_color: _background = "white",
):
    """\
    Create AnnData object for stLearn

    Parameters
    ----------
    count
        Pandas Dataframe of count matrix with rows as barcodes and columns as gene names
    spatial
        Pandas Dataframe of spatial location of cells/spots.
    library_id
        Identifier for the visium library. Can be modified when concatenating multiple adata objects.
    scale
        Set scale factor.
    quality
        Set quality that convert to stlearn to use. Store in anndata.obs['imagecol' & 'imagerow']
    spot_diameter_fullres
        Diameter of spot in full resolution
    background_color
        Color of the backgound. Only `black` or `white` is allowed.
    Returns
    -------
    AnnData
    """
    adata = AnnData(X=count)

    adata.obsm["spatial"] = spatial.values

    if scale == None:
        max_coor = np.max(adata.obsm["spatial"])
        scale = 2000 / max_coor

    adata.obs["imagecol"] = spatial["imagecol"].values * scale
    adata.obs["imagerow"] = spatial["imagerow"].values * scale

    if image_path != None:
        stlearn.add.image(
            adata,
            library_id=library_id,
            quality=quality,
            imgpath=image_path,
            scale=scale,
        )
    else:
        # Create image
        max_size = np.max([adata.obs["imagecol"].max(), adata.obs["imagerow"].max()])
        max_size = int(max_size + 0.1 * max_size)

        if background_color == "black":
            image = Image.new("RGBA", (max_size, max_size), (0, 0, 0, 0))
        else:
            image = Image.new("RGBA", (max_size, max_size), (255, 255, 255, 255))
        imgarr = np.array(image)

        # Create spatial dictionary
        adata.uns["spatial"] = {}
        adata.uns["spatial"][library_id] = {}
        adata.uns["spatial"][library_id]["images"] = {}
        adata.uns["spatial"][library_id]["images"][quality] = imgarr
        adata.uns["spatial"][library_id]["use_quality"] = quality
        adata.uns["spatial"][library_id]["scalefactors"] = {}
        adata.uns["spatial"][library_id]["scalefactors"][
            "tissue_" + quality + "_scalef"
        ] = scale
        adata.uns["spatial"][library_id]["scalefactors"][
            "spot_diameter_fullres"
        ] = spot_diameter_fullres

    return adata
